import json
from copy import deepcopy
from time import time
from unittest import IsolatedAsyncioTestCase

from .....cache.base import BaseCache
from .....cache.in_memory import InMemoryCache
from .....indy.holder import IndyHolder
from .....indy.issuer import IndyIssuer
from .....ledger.base import BaseLedger
from .....ledger.multiple_ledger.ledger_requests_executor import (
    IndyLedgerRequestsExecutor,
)
from .....messaging.credential_definitions.util import CRED_DEF_SENT_RECORD_TYPE
from .....messaging.decorators.thread_decorator import ThreadDecorator
from .....messaging.responder import BaseResponder, MockResponder
from .....multitenant.base import BaseMultitenantManager
from .....multitenant.manager import MultitenantManager
from .....storage.base import BaseStorage, StorageRecord
from .....storage.error import StorageNotFoundError
from .....tests import mock
from .....utils.testing import create_test_profile
from .. import manager as test_module
from ..manager import CredentialManager, CredentialManagerError
from ..messages.credential_ack import CredentialAck
from ..messages.credential_issue import CredentialIssue
from ..messages.credential_offer import CredentialOffer
from ..messages.credential_problem_report import CredentialProblemReport
from ..messages.credential_proposal import CredentialProposal
from ..messages.credential_request import CredentialRequest
from ..messages.inner.credential_preview import CredAttrSpec, CredentialPreview
from ..models.credential_exchange import V10CredentialExchange
from . import (
    CRED_DEF,
    CRED_DEF_ID,
    INDY_CRED,
    INDY_CRED_INFO,
    INDY_CRED_REQ,
    INDY_OFFER,
    REV_REG_DEF,
    REV_REG_ID,
    SCHEMA,
    SCHEMA_ID,
    TEST_DID,
)


class TestCredentialManager(IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.profile = await create_test_profile()

        self.ledger = mock.MagicMock(BaseLedger, autospec=True)
        self.ledger.get_schema = mock.CoroutineMock(return_value=SCHEMA)
        self.ledger.get_credential_definition = mock.CoroutineMock(return_value=CRED_DEF)
        self.ledger.get_revoc_reg_def = mock.CoroutineMock(return_value=REV_REG_DEF)
        self.ledger.credential_definition_id2schema_id = mock.CoroutineMock(
            return_value=SCHEMA_ID
        )
        self.profile.context.injector.bind_instance(BaseLedger, self.ledger)
        mock_executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        mock_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=(None, self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor, mock_executor
        )
        self.manager = CredentialManager(self.profile)
        assert self.manager.profile

    async def test_record_eq(self):
        same = [
            V10CredentialExchange(
                credential_exchange_id="dummy-0",
                thread_id="thread-0",
                credential_definition_id=CRED_DEF_ID,
                role=V10CredentialExchange.ROLE_ISSUER,
            )
        ] * 2
        diff = [
            V10CredentialExchange(
                credential_exchange_id="dummy-1",
                credential_definition_id=CRED_DEF_ID,
                role=V10CredentialExchange.ROLE_ISSUER,
            ),
            V10CredentialExchange(
                credential_exchange_id="dummy-0",
                thread_id="thread-1",
                credential_definition_id=CRED_DEF_ID,
                role=V10CredentialExchange.ROLE_ISSUER,
            ),
            V10CredentialExchange(
                credential_exchange_id="dummy-1",
                thread_id="thread-0",
                credential_definition_id=f"{CRED_DEF_ID}_distinct_tag",
                role=V10CredentialExchange.ROLE_ISSUER,
            ),
        ]

        for i in range(len(same) - 1):
            for j in range(i, len(same)):
                assert same[i] == same[j]

        for i in range(len(diff) - 1):
            for j in range(i, len(diff)):
                assert diff[i] == diff[j] if i == j else diff[i] != diff[j]

    async def test_prepare_send(self):
        connection_id = "test_conn_id"
        preview = CredentialPreview(
            attributes=(
                CredAttrSpec(name="legalName", value="value"),
                CredAttrSpec(name="jurisdictionId", value="value"),
                CredAttrSpec(name="incorporationDate", value="value"),
            )
        )
        proposal = CredentialProposal(
            credential_proposal=preview, cred_def_id=CRED_DEF_ID, schema_id=SCHEMA_ID
        )
        with mock.patch.object(
            self.manager, "create_offer", autospec=True
        ) as create_offer:
            create_offer.return_value = (mock.MagicMock(), mock.MagicMock())
            ret_exchange, _ = await self.manager.prepare_send(connection_id, proposal)
            create_offer.assert_called_once()
            assert ret_exchange is create_offer.return_value[0]
            arg_exchange = create_offer.call_args[1]["cred_ex_record"]
            assert arg_exchange.auto_issue
            assert arg_exchange.connection_id == connection_id
            assert arg_exchange.schema_id is None
            assert arg_exchange.credential_definition_id is None
            assert arg_exchange.role == V10CredentialExchange.ROLE_ISSUER
            assert arg_exchange.credential_proposal_dict == proposal

    async def test_create_proposal(self):
        connection_id = "test_conn_id"
        comment = "comment"
        preview = CredentialPreview(
            attributes=(
                CredAttrSpec(name="legalName", value="value"),
                CredAttrSpec(name="jurisdictionId", value="value"),
                CredAttrSpec(name="incorporationDate", value="value"),
            )
        )

        self.ledger.credential_definition_id2schema_id = mock.CoroutineMock(
            return_value=SCHEMA_ID
        )

        with mock.patch.object(V10CredentialExchange, "save", autospec=True) as save_ex:
            exchange: V10CredentialExchange = await self.manager.create_proposal(
                connection_id,
                auto_offer=True,
                comment=comment,
                credential_preview=preview,
                cred_def_id=CRED_DEF_ID,
            )
            save_ex.assert_called_once()

            await self.manager.create_proposal(
                connection_id,
                auto_offer=True,
                comment=comment,
                credential_preview=preview,
                cred_def_id=None,
            )  # OK to leave underspecified until offer

        proposal = exchange.credential_proposal_dict

        assert exchange.auto_offer
        assert exchange.connection_id == connection_id
        assert not exchange.credential_definition_id  # leave underspecified until offer
        assert not exchange.schema_id  # leave underspecified until offer
        assert exchange.thread_id == proposal._thread_id
        assert exchange.role == exchange.ROLE_HOLDER
        assert exchange.state == V10CredentialExchange.STATE_PROPOSAL_SENT

    async def test_create_proposal_no_preview(self):
        connection_id = "test_conn_id"
        comment = "comment"

        self.ledger.credential_definition_id2schema_id = mock.CoroutineMock(
            return_value=SCHEMA_ID
        )

        with mock.patch.object(V10CredentialExchange, "save", autospec=True) as save_ex:
            exchange: V10CredentialExchange = await self.manager.create_proposal(
                connection_id,
                auto_offer=True,
                comment=comment,
                credential_preview=None,
                cred_def_id=CRED_DEF_ID,
            )
            save_ex.assert_called_once()

        proposal = exchange.credential_proposal_dict

        assert exchange.auto_offer
        assert exchange.connection_id == connection_id
        assert not exchange.credential_definition_id  # leave underspecified until offer
        assert not exchange.schema_id  # leave underspecified until offer
        assert exchange.thread_id == proposal._thread_id
        assert exchange.role == exchange.ROLE_HOLDER
        assert exchange.state == V10CredentialExchange.STATE_PROPOSAL_SENT

    async def test_receive_proposal(self):
        connection_id = "test_conn_id"

        preview = CredentialPreview(
            attributes=(
                CredAttrSpec(name="legalName", value="value"),
                CredAttrSpec(name="jurisdictionId", value="value"),
                CredAttrSpec(name="incorporationDate", value="value"),
            )
        )

        with mock.patch.object(V10CredentialExchange, "save", autospec=True) as save_ex:
            proposal = CredentialProposal(
                credential_proposal=preview, cred_def_id=CRED_DEF_ID, schema_id=None
            )

            exchange = await self.manager.receive_proposal(proposal, connection_id)
            save_ex.assert_called_once()

            assert exchange.connection_id == connection_id
            assert exchange.credential_definition_id is None
            assert exchange.role == V10CredentialExchange.ROLE_ISSUER
            assert exchange.state == V10CredentialExchange.STATE_PROPOSAL_RECEIVED
            assert exchange.schema_id is None
            assert exchange.thread_id == proposal._thread_id

            ret_proposal: CredentialProposal = exchange.credential_proposal_dict
            attrs = ret_proposal.credential_proposal.attributes
            assert attrs == preview.attributes

            self.profile.context.message = CredentialProposal(
                credential_proposal=preview, cred_def_id=None, schema_id=None
            )
            await self.manager.receive_proposal(
                proposal, connection_id
            )  # OK to leave open until offer

    async def test_create_free_offer(self):
        comment = "comment"
        schema_id_parts = SCHEMA_ID.split(":")

        preview = CredentialPreview(
            attributes=(
                CredAttrSpec(name="legalName", value="value"),
                CredAttrSpec(name="jurisdictionId", value="value"),
                CredAttrSpec(name="incorporationDate", value="value"),
            )
        )
        proposal = CredentialProposal(
            credential_proposal=preview, cred_def_id=CRED_DEF_ID, schema_id=None
        )

        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            credential_definition_id=CRED_DEF_ID,
            role=V10CredentialExchange.ROLE_ISSUER,
            credential_proposal_dict=proposal.serialize(),
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        with mock.patch.object(V10CredentialExchange, "save", autospec=True) as save_ex:
            self.cache = InMemoryCache()
            self.profile.context.injector.bind_instance(BaseCache, self.cache)

            issuer = mock.MagicMock(IndyIssuer, autospec=True)
            issuer.create_credential_offer = mock.CoroutineMock(
                return_value=json.dumps(INDY_OFFER)
            )
            self.profile.context.injector.bind_instance(IndyIssuer, issuer)

            cred_def_record = StorageRecord(
                CRED_DEF_SENT_RECORD_TYPE,
                CRED_DEF_ID,
                {
                    "schema_id": SCHEMA_ID,
                    "schema_issuer_did": schema_id_parts[0],
                    "schema_name": schema_id_parts[-2],
                    "schema_version": schema_id_parts[-1],
                    "issuer_did": TEST_DID,
                    "cred_def_id": CRED_DEF_ID,
                    "epoch": str(int(time())),
                },
            )
            async with self.profile.session() as session:
                storage = session.inject(BaseStorage)
                await storage.add_record(cred_def_record)

            (ret_exchange, ret_offer) = await self.manager.create_offer(
                cred_ex_record=stored_exchange,
                counter_proposal=None,
                comment=comment,
            )
            assert ret_exchange is stored_exchange
            save_ex.assert_called_once()

            issuer.create_credential_offer.assert_called_once_with(CRED_DEF_ID)

            assert (
                stored_exchange.credential_exchange_id == ret_exchange._id
            )  # cover property
            assert stored_exchange.thread_id == ret_offer._thread_id
            assert stored_exchange.credential_definition_id == CRED_DEF_ID
            assert stored_exchange.role == V10CredentialExchange.ROLE_ISSUER
            assert stored_exchange.schema_id == SCHEMA_ID
            assert stored_exchange.state == V10CredentialExchange.STATE_OFFER_SENT
            assert stored_exchange._credential_offer.ser == INDY_OFFER

            await self.manager.create_offer(
                cred_ex_record=stored_exchange,
                counter_proposal=None,
                comment=comment,
            )  # once more to cover case where offer is available in cache

    async def test_create_free_offer_attr_mismatch(self):
        comment = "comment"
        schema_id_parts = SCHEMA_ID.split(":")

        preview = CredentialPreview(
            attributes=(
                CredAttrSpec(name="legal name", value="value"),
                CredAttrSpec(name="jurisdiction id", value="value"),
                CredAttrSpec(name="incorporation date", value="value"),
            )
        )
        proposal = CredentialProposal(
            credential_proposal=preview, cred_def_id=CRED_DEF_ID, schema_id=None
        )

        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            credential_definition_id=CRED_DEF_ID,
            role=V10CredentialExchange.ROLE_ISSUER,
            credential_proposal_dict=proposal.serialize(),
            new_with_id=True,
        )
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        with mock.patch.object(V10CredentialExchange, "save", autospec=True):
            self.cache = InMemoryCache()
            self.profile.context.injector.bind_instance(BaseCache, self.cache)

            issuer = mock.MagicMock(IndyIssuer, autospec=True)
            issuer.create_credential_offer = mock.CoroutineMock(
                return_value=json.dumps(INDY_OFFER)
            )
            self.profile.context.injector.bind_instance(IndyIssuer, issuer)

            cred_def_record = StorageRecord(
                CRED_DEF_SENT_RECORD_TYPE,
                CRED_DEF_ID,
                {
                    "schema_id": SCHEMA_ID,
                    "schema_issuer_did": schema_id_parts[0],
                    "schema_name": schema_id_parts[-2],
                    "schema_version": schema_id_parts[-1],
                    "issuer_did": TEST_DID,
                    "cred_def_id": CRED_DEF_ID,
                    "epoch": str(int(time())),
                },
            )
            async with self.profile.session() as session:
                storage = session.inject(BaseStorage)
                await storage.add_record(cred_def_record)

            with self.assertRaises(CredentialManagerError):
                await self.manager.create_offer(
                    cred_ex_record=stored_exchange,
                    counter_proposal=None,
                    comment=comment,
                )

    async def test_create_bound_offer(self):
        TEST_DID = "LjgpST2rjsoxYegQDRm7EL"
        schema_id_parts = SCHEMA_ID.split(":")
        comment = "comment"

        preview = CredentialPreview(
            attributes=(
                CredAttrSpec(name="legalName", value="value"),
                CredAttrSpec(name="jurisdictionId", value="value"),
                CredAttrSpec(name="incorporationDate", value="value"),
            )
        )
        proposal = CredentialProposal(credential_proposal=preview)
        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            credential_proposal_dict=proposal.serialize(),
            role=V10CredentialExchange.ROLE_ISSUER,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        with (
            mock.patch.object(V10CredentialExchange, "save", autospec=True) as save_ex,
            mock.patch.object(
                V10CredentialExchange, "get_cached_key", autospec=True
            ) as get_cached_key,
            mock.patch.object(V10CredentialExchange, "set_cached_key", autospec=True),
        ):
            get_cached_key.return_value = None
            issuer = mock.MagicMock(IndyIssuer, autospec=True)
            issuer.create_credential_offer = mock.CoroutineMock(
                return_value=json.dumps(INDY_OFFER)
            )
            self.profile.context.injector.bind_instance(IndyIssuer, issuer)

            cred_def_record = StorageRecord(
                CRED_DEF_SENT_RECORD_TYPE,
                CRED_DEF_ID,
                {
                    "schema_id": SCHEMA_ID,
                    "schema_issuer_did": schema_id_parts[0],
                    "schema_name": schema_id_parts[-2],
                    "schema_version": schema_id_parts[-1],
                    "issuer_did": TEST_DID,
                    "cred_def_id": CRED_DEF_ID,
                    "epoch": str(int(time())),
                },
            )
            async with self.profile.session() as session:
                storage = session.inject(BaseStorage)
                await storage.add_record(cred_def_record)

            (ret_exchange, ret_offer) = await self.manager.create_offer(
                cred_ex_record=stored_exchange,
                counter_proposal=None,
                comment=comment,
            )
            assert ret_exchange is stored_exchange
            save_ex.assert_called_once()

            issuer.create_credential_offer.assert_called_once_with(CRED_DEF_ID)

            assert stored_exchange.thread_id == ret_offer._thread_id
            assert stored_exchange.schema_id == SCHEMA_ID
            assert stored_exchange.credential_definition_id == CRED_DEF_ID
            assert stored_exchange.role == V10CredentialExchange.ROLE_ISSUER
            assert stored_exchange.state == V10CredentialExchange.STATE_OFFER_SENT
            assert stored_exchange._credential_offer.ser == INDY_OFFER

            # additionally check that credential preview was passed through
            assert ret_offer.credential_preview.attributes == preview.attributes

    async def test_create_bound_offer_no_cred_def(self):
        comment = "comment"

        preview = CredentialPreview(
            attributes=(
                CredAttrSpec(name="legalName", value="value"),
                CredAttrSpec(name="jurisdictionId", value="value"),
                CredAttrSpec(name="incorporationDate", value="value"),
            )
        )
        proposal = CredentialProposal(credential_proposal=preview)
        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            credential_proposal_dict=proposal.serialize(),
            role=V10CredentialExchange.ROLE_ISSUER,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        with (
            mock.patch.object(V10CredentialExchange, "save", autospec=True),
            mock.patch.object(
                V10CredentialExchange, "get_cached_key", autospec=True
            ) as get_cached_key,
            mock.patch.object(V10CredentialExchange, "set_cached_key", autospec=True),
        ):
            get_cached_key.return_value = None
            issuer = mock.MagicMock()
            issuer.create_credential_offer = mock.CoroutineMock(return_value=INDY_OFFER)
            self.profile.context.injector.bind_instance(IndyIssuer, issuer)

            with self.assertRaises(CredentialManagerError):
                await self.manager.create_offer(
                    cred_ex_record=stored_exchange,
                    counter_proposal=None,
                    comment=comment,
                )

    async def test_receive_offer_proposed(self):
        connection_id = "test_conn_id"
        thread_id = "thread-id"

        preview = CredentialPreview(
            attributes=(
                CredAttrSpec(name="legalName", value="value"),
                CredAttrSpec(name="jurisdictionId", value="value"),
                CredAttrSpec(name="incorporationDate", value="value"),
            )
        )
        proposal = CredentialProposal(credential_proposal=preview)

        offer = CredentialOffer(
            credential_preview=preview,
            offers_attach=[CredentialOffer.wrap_indy_offer(INDY_OFFER)],
        )
        offer.assign_thread_id(thread_id)

        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            credential_definition_id=CRED_DEF_ID,
            credential_proposal_dict=proposal.serialize(),
            initiator=V10CredentialExchange.INITIATOR_EXTERNAL,
            role=V10CredentialExchange.ROLE_HOLDER,
            state=V10CredentialExchange.STATE_PROPOSAL_SENT,
            schema_id=SCHEMA_ID,
            thread_id=thread_id,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        with (
            mock.patch.object(V10CredentialExchange, "save", autospec=True),
            mock.patch.object(
                V10CredentialExchange,
                "retrieve_by_connection_and_thread",
                mock.CoroutineMock(return_value=stored_exchange),
            ),
        ):
            exchange = await self.manager.receive_offer(offer, connection_id)

            assert exchange.connection_id == connection_id
            assert exchange.credential_definition_id == CRED_DEF_ID
            assert exchange.schema_id == SCHEMA_ID
            assert exchange.thread_id == offer._thread_id
            assert exchange.role == V10CredentialExchange.ROLE_HOLDER
            assert exchange.state == V10CredentialExchange.STATE_OFFER_RECEIVED
            assert exchange._credential_offer.ser == INDY_OFFER
            assert exchange.credential_offer_dict == offer

            proposal = exchange.credential_proposal_dict
            assert proposal.credential_proposal.attributes == preview.attributes

    async def test_receive_free_offer(self):
        connection_id = "test_conn_id"
        preview = CredentialPreview(
            attributes=(
                CredAttrSpec(name="legalName", value="value"),
                CredAttrSpec(name="jurisdictionId", value="value"),
                CredAttrSpec(name="incorporationDate", value="value"),
            )
        )

        offer = CredentialOffer(
            credential_preview=preview,
            offers_attach=[CredentialOffer.wrap_indy_offer(INDY_OFFER)],
        )
        self.profile.context.message = offer
        self.profile.context.connection_record = mock.MagicMock()
        self.profile.context.connection_record.connection_id = connection_id

        with (
            mock.patch.object(V10CredentialExchange, "save", autospec=True),
            mock.patch.object(
                V10CredentialExchange,
                "retrieve_by_connection_and_thread",
                mock.CoroutineMock(side_effect=StorageNotFoundError),
            ),
        ):
            exchange = await self.manager.receive_offer(offer, connection_id)

            assert exchange.connection_id == connection_id
            assert exchange.credential_definition_id == CRED_DEF_ID
            assert exchange.schema_id == SCHEMA_ID
            assert exchange.thread_id == offer._thread_id
            assert exchange.role == V10CredentialExchange.ROLE_HOLDER
            assert exchange.state == V10CredentialExchange.STATE_OFFER_RECEIVED
            assert exchange._credential_offer.ser == INDY_OFFER
            assert exchange.credential_proposal_dict
            assert exchange.credential_offer_dict == offer

    async def test_create_request(self):
        connection_id = "test_conn_id"
        thread_id = "thread-id"
        holder_did = "did"

        credential_offer_dict = CredentialOffer(
            "thread-id",
        )
        credential_offer_dict._thread = ThreadDecorator(pthid="some-pthid")

        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            credential_definition_id=CRED_DEF_ID,
            credential_offer=INDY_OFFER,
            initiator=V10CredentialExchange.INITIATOR_SELF,
            role=V10CredentialExchange.ROLE_HOLDER,
            state=V10CredentialExchange.STATE_OFFER_RECEIVED,
            credential_offer_dict=credential_offer_dict,
            schema_id=SCHEMA_ID,
            thread_id=thread_id,
            new_with_id=True,
        )

        async with self.profile.session() as session:
            await stored_exchange.save(session)

        self.cache = InMemoryCache()
        self.profile.context.injector.bind_instance(BaseCache, self.cache)

        with mock.patch.object(V10CredentialExchange, "save", autospec=True):
            cred_def = {"cred": "def"}
            self.ledger.get_credential_definition = mock.CoroutineMock(
                return_value=cred_def
            )

            cred_req_meta = {}
            holder = mock.MagicMock(IndyHolder, autospec=True)
            holder.create_credential_request = mock.CoroutineMock(
                return_value=(json.dumps(INDY_CRED_REQ), json.dumps(cred_req_meta))
            )
            self.profile.context.injector.bind_instance(IndyHolder, holder)

            ret_exchange, ret_request = await self.manager.create_request(
                stored_exchange, holder_did
            )

            holder.create_credential_request.assert_called_once_with(
                INDY_OFFER, cred_def, holder_did
            )

            assert ret_request.indy_cred_req() == INDY_CRED_REQ
            assert ret_request._thread_id == thread_id

            assert ret_exchange.state == V10CredentialExchange.STATE_REQUEST_SENT

            # cover case with request in cache
            stored_exchange.credential_request = None
            stored_exchange.state = V10CredentialExchange.STATE_OFFER_RECEIVED
            await self.manager.create_request(stored_exchange, holder_did)

            # cover case with existing cred req
            (
                ret_existing_exchange,
                ret_existing_request,
            ) = await self.manager.create_request(ret_exchange, holder_did)
            assert ret_existing_exchange == ret_exchange
            assert ret_existing_request._thread_id == thread_id
            assert ret_existing_request._thread.pthid == "some-pthid"

    async def test_create_request_no_cache(self):
        connection_id = "test_conn_id"
        thread_id = "thread-id"
        holder_did = "did"

        credential_offer_dict = CredentialOffer(
            "thread-id",
        )
        credential_offer_dict._thread = ThreadDecorator(pthid="some-pthid")

        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            credential_definition_id=CRED_DEF_ID,
            credential_offer=INDY_OFFER,
            credential_offer_dict=credential_offer_dict,
            initiator=V10CredentialExchange.INITIATOR_SELF,
            role=V10CredentialExchange.ROLE_HOLDER,
            state=V10CredentialExchange.STATE_OFFER_RECEIVED,
            schema_id=SCHEMA_ID,
            thread_id=thread_id,
            new_with_id=True,
        )
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        with mock.patch.object(V10CredentialExchange, "save", autospec=True):
            cred_def = {"cred": "def"}
            self.ledger.get_credential_definition = mock.CoroutineMock(
                return_value=cred_def
            )

            cred_req_meta = {}
            holder = mock.MagicMock(IndyHolder, autospec=True)
            holder.create_credential_request = mock.CoroutineMock(
                return_value=(json.dumps(INDY_CRED_REQ), json.dumps(cred_req_meta))
            )
            self.profile.context.injector.bind_instance(IndyHolder, holder)

            ret_exchange, ret_request = await self.manager.create_request(
                stored_exchange, holder_did
            )

            holder.create_credential_request.assert_called_once_with(
                INDY_OFFER, cred_def, holder_did
            )

            assert ret_request.indy_cred_req() == INDY_CRED_REQ
            assert ret_request._thread_id == thread_id
            assert ret_request._thread.pthid == "some-pthid"

            assert ret_exchange.state == V10CredentialExchange.STATE_REQUEST_SENT

    async def test_create_request_bad_state(self):
        connection_id = "test_conn_id"
        thread_id = "thread-id"
        holder_did = "did"

        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            credential_definition_id=CRED_DEF_ID,
            credential_offer=INDY_OFFER,
            initiator=V10CredentialExchange.INITIATOR_SELF,
            role=V10CredentialExchange.ROLE_HOLDER,
            state=V10CredentialExchange.STATE_PROPOSAL_SENT,
            schema_id=SCHEMA_ID,
            thread_id=thread_id,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        with self.assertRaises(CredentialManagerError):
            await self.manager.create_request(stored_exchange, holder_did)

    async def test_receive_request(self):
        mock_conn = mock.MagicMock(connection_id="test_conn_id")

        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=mock_conn.connection_id,
            initiator=V10CredentialExchange.INITIATOR_EXTERNAL,
            role=V10CredentialExchange.ROLE_ISSUER,
            state=V10CredentialExchange.STATE_OFFER_SENT,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

            request = CredentialRequest(
                requests_attach=[CredentialRequest.wrap_indy_cred_req(INDY_CRED_REQ)]
            )

            with (
                mock.patch.object(
                    V10CredentialExchange, "save", autospec=True
                ) as save_ex,
                mock.patch.object(
                    V10CredentialExchange,
                    "retrieve_by_connection_and_thread",
                    mock.CoroutineMock(return_value=stored_exchange),
                ) as retrieve_ex,
            ):
                exchange = await self.manager.receive_request(request, mock_conn, None)

            retrieve_ex.assert_called()
            save_ex.assert_called_once()

            assert exchange.state == V10CredentialExchange.STATE_REQUEST_RECEIVED
            assert exchange._credential_request.ser == INDY_CRED_REQ

    async def test_receive_request_no_connection_cred_request(self):
        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            initiator=V10CredentialExchange.INITIATOR_EXTERNAL,
            role=V10CredentialExchange.ROLE_ISSUER,
            state=V10CredentialExchange.STATE_OFFER_SENT,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        request = CredentialRequest(
            requests_attach=[CredentialRequest.wrap_indy_cred_req(INDY_CRED_REQ)]
        )

        mock_conn = mock.MagicMock(
            connection_id="test_conn_id",
        )
        mock_oob = mock.MagicMock()

        with (
            mock.patch.object(V10CredentialExchange, "save", autospec=True) as mock_save,
            mock.patch.object(
                V10CredentialExchange,
                "retrieve_by_connection_and_thread",
                mock.CoroutineMock(),
            ) as mock_retrieve,
        ):
            mock_retrieve.return_value = stored_exchange
            cx_rec = await self.manager.receive_request(request, mock_conn, mock_oob)

            mock_retrieve.assert_called()
            mock_save.assert_called_once()
            assert cx_rec.state == V10CredentialExchange.STATE_REQUEST_RECEIVED
            assert cx_rec._credential_request.ser == INDY_CRED_REQ
            assert cx_rec.connection_id == "test_conn_id"

    async def test_receive_request_no_cred_ex_with_offer_found(self):
        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            initiator=V10CredentialExchange.INITIATOR_EXTERNAL,
            role=V10CredentialExchange.ROLE_ISSUER,
            state=V10CredentialExchange.STATE_OFFER_SENT,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        request = CredentialRequest(
            requests_attach=[CredentialRequest.wrap_indy_cred_req(INDY_CRED_REQ)]
        )

        mock_conn = mock.MagicMock(
            connection_id="test_conn_id",
        )

        with (
            mock.patch.object(V10CredentialExchange, "save", autospec=True),
            mock.patch.object(
                V10CredentialExchange,
                "retrieve_by_connection_and_thread",
                mock.CoroutineMock(),
            ) as mock_retrieve,
        ):
            mock_retrieve.side_effect = (StorageNotFoundError(),)
            with self.assertRaises(StorageNotFoundError):
                await self.manager.receive_request(request, mock_conn, None)

                mock_retrieve.assert_called()

    async def test_issue_credential_revocable(self):
        connection_id = "test_conn_id"
        comment = "comment"
        thread_id = "thread-id"

        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            credential_definition_id=CRED_DEF_ID,
            credential_offer=INDY_OFFER,
            credential_request=INDY_CRED_REQ,
            credential_proposal_dict=CredentialProposal(
                credential_proposal=CredentialPreview.deserialize(
                    {"attributes": [{"name": "attr", "value": "value"}]}
                ),
                cred_def_id=CRED_DEF_ID,
                schema_id=SCHEMA_ID,
            ).serialize(),
            initiator=V10CredentialExchange.INITIATOR_SELF,
            role=V10CredentialExchange.ROLE_ISSUER,
            state=V10CredentialExchange.STATE_REQUEST_RECEIVED,
            thread_id=thread_id,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        issuer = mock.MagicMock(IndyIssuer, autospec=True)
        cred = {"indy": "credential"}
        cred_rev_id = "1000"
        issuer.create_credential = mock.CoroutineMock(
            return_value=(json.dumps(cred), cred_rev_id)
        )
        self.profile.context.injector.bind_instance(IndyIssuer, issuer)

        with (
            mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc,
            mock.patch.object(V10CredentialExchange, "save", autospec=True) as save_ex,
        ):
            revoc.return_value.get_or_create_active_registry = mock.CoroutineMock(
                return_value=(
                    mock.MagicMock(  # active_rev_reg_rec
                        revoc_reg_id=REV_REG_ID,
                    ),
                    mock.MagicMock(  # rev_reg
                        registry_id=REV_REG_ID,
                        tails_local_path="dummy-path",
                        get_or_fetch_local_tails_path=mock.CoroutineMock(),
                        max_creds=10,
                    ),
                )
            )
            (ret_exchange, ret_cred_issue) = await self.manager.issue_credential(
                stored_exchange, comment=comment, retries=1
            )

            save_ex.assert_called_once()

            issuer.create_credential.assert_called()

            assert ret_exchange._credential.ser == cred
            assert ret_cred_issue.indy_credential() == cred
            assert ret_exchange.state == V10CredentialExchange.STATE_ISSUED
            assert ret_cred_issue._thread_id == thread_id

            # cover case with existing cred
            (
                ret_existing_exchange,
                ret_existing_cred,
            ) = await self.manager.issue_credential(
                ret_exchange, comment=comment, retries=0
            )
            assert ret_existing_exchange == ret_exchange
            assert ret_existing_cred._thread_id == thread_id

    async def test_issue_credential_non_revocable(self):
        CRED_DEF_NR = deepcopy(CRED_DEF)
        CRED_DEF_NR["value"]["revocation"] = None
        connection_id = "test_conn_id"
        comment = "comment"
        cred_values = {"attr": "value"}
        thread_id = "thread-id"
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            credential_definition_id=CRED_DEF_ID,
            credential_offer=INDY_OFFER,
            credential_request=INDY_CRED_REQ,
            credential_proposal_dict=CredentialProposal(
                credential_proposal=CredentialPreview.deserialize(
                    {"attributes": [{"name": "attr", "value": "value"}]}
                ),
                cred_def_id=CRED_DEF_ID,
                schema_id=SCHEMA_ID,
            ).serialize(),
            initiator=V10CredentialExchange.INITIATOR_SELF,
            role=V10CredentialExchange.ROLE_ISSUER,
            state=V10CredentialExchange.STATE_REQUEST_RECEIVED,
            thread_id=thread_id,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        issuer = mock.MagicMock(IndyIssuer, autospec=True)
        cred = {"indy": "credential"}
        issuer.create_credential = mock.CoroutineMock(
            return_value=(json.dumps(cred), None)
        )
        self.profile.context.injector.bind_instance(IndyIssuer, issuer)

        self.ledger = mock.MagicMock(BaseLedger, autospec=True)
        self.ledger.get_schema = mock.CoroutineMock(return_value=SCHEMA)
        self.ledger.get_credential_definition = mock.CoroutineMock(
            return_value=CRED_DEF_NR
        )
        self.ledger.__aenter__ = mock.CoroutineMock(return_value=self.ledger)
        self.profile.context.injector.clear_binding(BaseLedger)
        self.profile.context.injector.bind_instance(BaseLedger, self.ledger)
        with (
            mock.patch.object(V10CredentialExchange, "save", autospec=True) as save_ex,
            mock.patch.object(
                IndyLedgerRequestsExecutor,
                "get_ledger_for_identifier",
                mock.CoroutineMock(return_value=("test_ledger_id", self.ledger)),
            ),
        ):
            (ret_exchange, ret_cred_issue) = await self.manager.issue_credential(
                stored_exchange, comment=comment, retries=0
            )

            save_ex.assert_called_once()

            issuer.create_credential.assert_called_once_with(
                SCHEMA,
                INDY_OFFER,
                INDY_CRED_REQ,
                cred_values,
                None,
                None,
            )

            assert ret_exchange._credential.ser == cred
            assert ret_cred_issue.indy_credential() == cred
            assert ret_exchange.state == V10CredentialExchange.STATE_ISSUED
            assert ret_cred_issue._thread_id == thread_id

    async def test_issue_credential_fills_rr(self):
        connection_id = "test_conn_id"
        comment = "comment"
        cred_values = {"attr": "value"}
        thread_id = "thread-id"

        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            credential_definition_id=CRED_DEF_ID,
            credential_offer=INDY_OFFER,
            credential_request=INDY_CRED_REQ,
            credential_proposal_dict=CredentialProposal(
                credential_proposal=CredentialPreview.deserialize(
                    {"attributes": [{"name": "attr", "value": "value"}]}
                ),
                cred_def_id=CRED_DEF_ID,
                schema_id=SCHEMA_ID,
            ).serialize(),
            initiator=V10CredentialExchange.INITIATOR_SELF,
            role=V10CredentialExchange.ROLE_ISSUER,
            state=V10CredentialExchange.STATE_REQUEST_RECEIVED,
            thread_id=thread_id,
            revocation_id="1000",
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        issuer = mock.MagicMock(IndyIssuer, autospec=True)
        cred = {"indy": "credential"}
        issuer.create_credential = mock.CoroutineMock(
            return_value=(json.dumps(cred), stored_exchange.revocation_id)
        )
        self.profile.context.injector.bind_instance(IndyIssuer, issuer)

        with (
            mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc,
            mock.patch.object(V10CredentialExchange, "save", autospec=True) as save_ex,
        ):
            revoc.return_value = mock.MagicMock(
                get_or_create_active_registry=(
                    mock.CoroutineMock(
                        return_value=(
                            mock.MagicMock(  # active_rev_reg_rec
                                revoc_reg_id=REV_REG_ID,
                                set_state=mock.CoroutineMock(),
                            ),
                            mock.MagicMock(  # rev_reg
                                registry_id=REV_REG_ID,
                                tails_local_path="dummy-path",
                                max_creds=1000,
                                get_or_fetch_local_tails_path=(mock.CoroutineMock()),
                            ),
                        )
                    )
                ),
                handle_full_registry=mock.CoroutineMock(),
            )
            (ret_exchange, ret_cred_issue) = await self.manager.issue_credential(
                stored_exchange, comment=comment, retries=0
            )

            save_ex.assert_called_once()

            issuer.create_credential.assert_called_once_with(
                SCHEMA,
                INDY_OFFER,
                INDY_CRED_REQ,
                cred_values,
                REV_REG_ID,
                "dummy-path",
            )

            revoc.return_value.handle_full_registry.assert_awaited_once_with(REV_REG_ID)

            assert ret_exchange._credential.ser == cred
            assert ret_cred_issue.indy_credential() == cred
            assert ret_exchange.state == V10CredentialExchange.STATE_ISSUED
            assert ret_cred_issue._thread_id == thread_id

    async def test_issue_credential_request_bad_state(self):
        connection_id = "test_conn_id"
        thread_id = "thread-id"

        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            credential_definition_id=CRED_DEF_ID,
            credential_offer=INDY_OFFER,
            initiator=V10CredentialExchange.INITIATOR_SELF,
            role=V10CredentialExchange.ROLE_HOLDER,
            state=V10CredentialExchange.STATE_PROPOSAL_SENT,
            schema_id=SCHEMA_ID,
            thread_id=thread_id,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        with self.assertRaises(CredentialManagerError):
            await self.manager.issue_credential(stored_exchange)

    async def test_issue_credential_no_active_rr_no_retries(self):
        connection_id = "test_conn_id"
        comment = "comment"
        thread_id = "thread-id"

        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            credential_definition_id=CRED_DEF_ID,
            credential_offer=INDY_OFFER,
            credential_request=INDY_CRED_REQ,
            credential_proposal_dict=CredentialProposal(
                credential_proposal=CredentialPreview.deserialize(
                    {"attributes": [{"name": "attr", "value": "value"}]}
                ),
                cred_def_id=CRED_DEF_ID,
                schema_id=SCHEMA_ID,
            ).serialize(),
            initiator=V10CredentialExchange.INITIATOR_SELF,
            role=V10CredentialExchange.ROLE_ISSUER,
            state=V10CredentialExchange.STATE_REQUEST_RECEIVED,
            thread_id=thread_id,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        issuer = mock.MagicMock(IndyIssuer, autospec=True)
        cred = {"indy": "credential"}
        cred_rev_id = "1"
        issuer.create_credential = mock.CoroutineMock(
            return_value=(json.dumps(cred), cred_rev_id)
        )
        self.profile.context.injector.bind_instance(IndyIssuer, issuer)
        executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=("test_ledger_id", self.ledger)
        )
        self.profile.context.injector.bind_instance(IndyLedgerRequestsExecutor, executor)
        with mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc:
            revoc.return_value.get_or_create_active_registry = mock.CoroutineMock(
                side_effect=[
                    None,
                    (
                        mock.MagicMock(  # active_rev_reg_rec
                            revoc_reg_id=REV_REG_ID,
                            set_state=mock.CoroutineMock(),
                        ),
                        mock.MagicMock(  # rev_reg
                            tails_local_path="dummy-path",
                            get_or_fetch_local_tails_path=(mock.CoroutineMock()),
                        ),
                    ),
                ]
            )
            with self.assertRaises(CredentialManagerError):
                await self.manager.issue_credential(
                    stored_exchange, comment=comment, retries=0
                )

    async def test_issue_credential_no_active_rr_retry(self):
        connection_id = "test_conn_id"
        comment = "comment"
        thread_id = "thread-id"

        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            credential_definition_id=CRED_DEF_ID,
            credential_offer=INDY_OFFER,
            credential_request=INDY_CRED_REQ,
            credential_proposal_dict=CredentialProposal(
                credential_proposal=CredentialPreview.deserialize(
                    {"attributes": [{"name": "attr", "value": "value"}]}
                ),
                cred_def_id=CRED_DEF_ID,
                schema_id=SCHEMA_ID,
            ).serialize(),
            initiator=V10CredentialExchange.INITIATOR_SELF,
            role=V10CredentialExchange.ROLE_ISSUER,
            state=V10CredentialExchange.STATE_REQUEST_RECEIVED,
            thread_id=thread_id,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        issuer = mock.MagicMock(IndyIssuer, autospec=True)
        cred = {"indy": "credential"}
        cred_rev_id = "1"
        issuer.create_credential = mock.CoroutineMock(
            return_value=(json.dumps(cred), cred_rev_id)
        )
        self.profile.context.injector.bind_instance(IndyIssuer, issuer)
        executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=("test_ledger_id", self.ledger)
        )
        self.profile.context.injector.bind_instance(IndyLedgerRequestsExecutor, executor)
        with mock.patch.object(test_module, "IndyRevocation", autospec=True) as revoc:
            revoc.return_value.get_or_create_active_registry = mock.CoroutineMock(
                return_value=None
            )
            with self.assertRaises(CredentialManagerError):
                await self.manager.issue_credential(
                    stored_exchange, comment=comment, retries=1
                )

    async def test_receive_credential(self):
        connection_id = "test_conn_id"

        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            initiator=V10CredentialExchange.INITIATOR_EXTERNAL,
            role=V10CredentialExchange.ROLE_HOLDER,
            state=V10CredentialExchange.STATE_REQUEST_SENT,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        issue = CredentialIssue(
            credentials_attach=[CredentialIssue.wrap_indy_credential(INDY_CRED)]
        )

        with (
            mock.patch.object(V10CredentialExchange, "save", autospec=True) as save_ex,
            mock.patch.object(
                V10CredentialExchange,
                "retrieve_by_connection_and_thread",
                mock.CoroutineMock(return_value=stored_exchange),
            ) as retrieve_ex,
        ):
            exchange = await self.manager.receive_credential(issue, connection_id)

            assert retrieve_ex.call_args.args[1] == connection_id
            assert retrieve_ex.call_args.args[2] == issue._thread_id
            assert (
                retrieve_ex.call_args.kwargs["role"] == V10CredentialExchange.ROLE_HOLDER
            )
            assert retrieve_ex.call_args.kwargs["for_update"] is True
            save_ex.assert_called_once()

            assert exchange._raw_credential.ser == INDY_CRED
            assert exchange.state == V10CredentialExchange.STATE_CREDENTIAL_RECEIVED

    async def test_store_credential(self):
        connection_id = "test_conn_id"
        cred_req_meta = {"req": "meta"}
        thread_id = "thread-id"

        preview = CredentialPreview(
            attributes=(
                CredAttrSpec(
                    name="legalName", value="value", mime_type="text/plain;lang=en-ca"
                ),
                CredAttrSpec(name="jurisdictionId", value="value"),
                CredAttrSpec(name="incorporationDate", value="value"),
            )
        )
        proposal = CredentialProposal(
            credential_proposal=preview, cred_def_id=CRED_DEF_ID, schema_id=SCHEMA_ID
        )

        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            credential_definition_id=CRED_DEF_ID,
            credential_request_metadata=cred_req_meta,
            credential_proposal_dict=proposal,
            raw_credential=INDY_CRED,
            initiator=V10CredentialExchange.INITIATOR_EXTERNAL,
            role=V10CredentialExchange.ROLE_HOLDER,
            state=V10CredentialExchange.STATE_CREDENTIAL_RECEIVED,
            thread_id=thread_id,
            auto_remove=True,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        cred_id = "cred-id"
        holder = mock.MagicMock(IndyHolder, autospec=True)
        holder.store_credential = mock.CoroutineMock(return_value=cred_id)
        holder.get_credential = mock.CoroutineMock(
            return_value=json.dumps(INDY_CRED_INFO)
        )
        self.profile.context.injector.bind_instance(IndyHolder, holder)
        executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=("test_ledger_id", self.ledger)
        )
        self.profile.context.injector.bind_instance(IndyLedgerRequestsExecutor, executor)
        with (
            mock.patch.object(
                test_module, "RevocationRegistry", autospec=True
            ) as mock_rev_reg,
            mock.patch.object(V10CredentialExchange, "save", autospec=True) as save_ex,
            mock.patch.object(V10CredentialExchange, "delete_record", autospec=True),
        ):
            mock_rev_reg.from_definition = mock.MagicMock(
                return_value=mock.MagicMock(
                    get_or_fetch_local_tails_path=mock.CoroutineMock()
                )
            )
            ret_exchange = await self.manager.store_credential(
                stored_exchange, credential_id=cred_id
            )

            save_ex.assert_called_once()

            self.ledger.get_credential_definition.assert_called_once_with(CRED_DEF_ID)

            holder.store_credential.assert_called_once_with(
                CRED_DEF,
                INDY_CRED,
                cred_req_meta,
                {"legalName": "text/plain;lang=en-ca"},
                credential_id=cred_id,
                rev_reg_def=REV_REG_DEF,
            )

            holder.get_credential.assert_called_once_with(cred_id)

            assert ret_exchange.credential_id == cred_id
            assert ret_exchange._credential.ser == INDY_CRED_INFO
            assert ret_exchange.state == V10CredentialExchange.STATE_CREDENTIAL_RECEIVED

    async def test_store_credential_bad_state(self):
        connection_id = "test_conn_id"
        cred_req_meta = {"req": "meta"}
        thread_id = "thread-id"

        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            credential_definition_id=CRED_DEF_ID,
            credential_request_metadata=cred_req_meta,
            credential_proposal_dict=None,
            raw_credential=INDY_CRED,
            initiator=V10CredentialExchange.INITIATOR_EXTERNAL,
            role=V10CredentialExchange.ROLE_HOLDER,
            state=V10CredentialExchange.STATE_OFFER_RECEIVED,
            thread_id=thread_id,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)
        cred_id = "cred-id"

        with self.assertRaises(CredentialManagerError):
            await self.manager.store_credential(stored_exchange, credential_id=cred_id)

    async def test_store_credential_no_preview(self):
        connection_id = "test_conn_id"
        cred_req_meta = {"req": "meta"}
        thread_id = "thread-id"
        self.profile.context.injector.bind_instance(
            BaseMultitenantManager,
            mock.MagicMock(MultitenantManager, autospec=True),
        )
        cred_no_rev = {**INDY_CRED}
        cred_no_rev["rev_reg_id"] = None
        cred_no_rev["rev_reg"] = None
        cred_no_rev["witness"] = None
        cred_info_no_rev = {**INDY_CRED_INFO}
        cred_info_no_rev["rev_reg_id"] = None
        cred_info_no_rev["cred_rev_id"] = None
        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            credential_definition_id=CRED_DEF_ID,
            credential_request_metadata=cred_req_meta,
            credential_proposal_dict=None,
            raw_credential=cred_no_rev,
            initiator=V10CredentialExchange.INITIATOR_EXTERNAL,
            role=V10CredentialExchange.ROLE_HOLDER,
            state=V10CredentialExchange.STATE_CREDENTIAL_RECEIVED,
            thread_id=thread_id,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        cred_def = mock.MagicMock()
        self.ledger.get_credential_definition = mock.CoroutineMock(return_value=cred_def)

        cred_id = "cred-id"
        holder = mock.MagicMock(IndyHolder, autospec=True)
        holder.store_credential = mock.CoroutineMock(return_value=cred_id)
        holder.get_credential = mock.CoroutineMock(
            return_value=json.dumps(cred_info_no_rev)
        )
        self.profile.context.injector.bind_instance(IndyHolder, holder)
        executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=("test_ledger_id", self.ledger)
        )
        self.profile.context.injector.bind_instance(IndyLedgerRequestsExecutor, executor)
        with (
            mock.patch.object(V10CredentialExchange, "save", autospec=True) as save_ex,
            mock.patch.object(V10CredentialExchange, "delete_record", autospec=True),
        ):
            ret_exchange = await self.manager.store_credential(stored_exchange)

            save_ex.assert_called_once()

            self.ledger.get_credential_definition.assert_called_once_with(CRED_DEF_ID)

            holder.store_credential.assert_called_once_with(
                cred_def,
                cred_no_rev,
                cred_req_meta,
                None,
                credential_id=None,
                rev_reg_def=None,
            )

            holder.get_credential.assert_called_once_with(cred_id)

            assert ret_exchange.credential_id == cred_id
            assert ret_exchange._credential.ser == cred_info_no_rev
            assert ret_exchange.state == V10CredentialExchange.STATE_CREDENTIAL_RECEIVED

    async def test_store_credential_holder_store_indy_error(self):
        connection_id = "test_conn_id"
        cred_req_meta = {"req": "meta"}
        thread_id = "thread-id"

        cred_no_rev = {**INDY_CRED}
        cred_no_rev["rev_reg_id"] = None
        cred_no_rev["rev_reg"] = None
        cred_no_rev["witness"] = None
        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            credential_definition_id=CRED_DEF_ID,
            credential_request_metadata=cred_req_meta,
            credential_proposal_dict=None,
            raw_credential=cred_no_rev,
            initiator=V10CredentialExchange.INITIATOR_EXTERNAL,
            role=V10CredentialExchange.ROLE_HOLDER,
            state=V10CredentialExchange.STATE_CREDENTIAL_RECEIVED,
            thread_id=thread_id,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        cred_def = mock.MagicMock()
        self.ledger.get_credential_definition = mock.CoroutineMock(return_value=cred_def)

        cred_id = "cred-id"
        holder = mock.MagicMock(IndyHolder, autospec=True)
        holder.store_credential = mock.CoroutineMock(
            side_effect=test_module.IndyHolderError("Problem", {"message": "Nope"})
        )
        self.profile.context.injector.bind_instance(IndyHolder, holder)

        mock_executor = mock.MagicMock(IndyLedgerRequestsExecutor, autospec=True)
        mock_executor.get_ledger_for_identifier = mock.CoroutineMock(
            return_value=("test_ledger_id", self.ledger)
        )
        self.profile.context.injector.bind_instance(
            IndyLedgerRequestsExecutor, mock_executor
        )
        with self.assertRaises(test_module.IndyHolderError):
            await self.manager.store_credential(
                cred_ex_record=stored_exchange, credential_id=cred_id
            )

    async def test_send_credential_ack(self):
        connection_id = "connection-id"
        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            initiator=V10CredentialExchange.INITIATOR_SELF,
            state=V10CredentialExchange.STATE_CREDENTIAL_RECEIVED,
            thread_id="thid",
            parent_thread_id="pthid",
            role=V10CredentialExchange.ROLE_ISSUER,
            trace=False,
            auto_remove=True,
            new_with_id=True,
        )

        async with self.profile.session() as session:
            await stored_exchange.save(session)

        with (
            mock.patch.object(V10CredentialExchange, "save", autospec=True),
            mock.patch.object(
                V10CredentialExchange, "delete_record", autospec=True
            ) as mock_delete_ex,
            mock.patch.object(
                test_module.LOGGER, "exception", mock.MagicMock()
            ) as mock_log_exception,
            mock.patch.object(
                test_module.LOGGER, "warning", mock.MagicMock()
            ) as mock_log_warning,
        ):
            mock_delete_ex.side_effect = test_module.StorageError()
            (exch, ack) = await self.manager.send_credential_ack(stored_exchange)
            assert ack._thread
            mock_log_exception.assert_called_once()  # cover exception log-and-continue
            mock_log_warning.assert_called_once()  # no BaseResponder
            assert exch.state == V10CredentialExchange.STATE_ACKED

            mock_responder = MockResponder()  # cover with responder
            self.profile.context.injector.bind_instance(BaseResponder, mock_responder)
            (exch, ack) = await self.manager.send_credential_ack(stored_exchange)
            assert ack._thread
            assert exch.state == V10CredentialExchange.STATE_ACKED

    async def test_receive_credential_ack(self):
        connection_id = "connection-id"
        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            initiator=V10CredentialExchange.INITIATOR_SELF,
            role=V10CredentialExchange.ROLE_ISSUER,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)

        ack = CredentialAck()

        with (
            mock.patch.object(V10CredentialExchange, "save", autospec=True) as save_ex,
            mock.patch.object(
                V10CredentialExchange, "delete_record", autospec=True
            ) as delete_ex,
            mock.patch.object(
                V10CredentialExchange,
                "retrieve_by_connection_and_thread",
                mock.CoroutineMock(),
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = stored_exchange
            ret_exchange = await self.manager.receive_credential_ack(ack, connection_id)

            assert retrieve_ex.call_args.args[1] == connection_id
            assert retrieve_ex.call_args.args[2] == ack._thread_id
            assert (
                retrieve_ex.call_args.kwargs["role"] == V10CredentialExchange.ROLE_ISSUER
            )
            assert retrieve_ex.call_args.kwargs["for_update"] is True
            save_ex.assert_called_once()

            assert ret_exchange.state == V10CredentialExchange.STATE_ACKED
            delete_ex.assert_called_once()

    async def test_receive_problem_report(self):
        connection_id = "connection-id"
        stored_exchange = V10CredentialExchange(
            credential_exchange_id="dummy-cxid",
            connection_id=connection_id,
            initiator=V10CredentialExchange.INITIATOR_SELF,
            role=V10CredentialExchange.ROLE_ISSUER,
            new_with_id=True,
        )
        async with self.profile.session() as session:
            await stored_exchange.save(session)
        problem = CredentialProblemReport(
            description={
                "code": test_module.ProblemReportReason.ISSUANCE_ABANDONED.value,
                "en": "Insufficient privilege",
            }
        )

        with (
            mock.patch.object(V10CredentialExchange, "save", autospec=True) as save_ex,
            mock.patch.object(
                V10CredentialExchange,
                "retrieve_by_connection_and_thread",
                mock.CoroutineMock(),
            ) as retrieve_ex,
        ):
            retrieve_ex.return_value = stored_exchange

            ret_exchange = await self.manager.receive_problem_report(
                problem, connection_id
            )
            assert retrieve_ex.call_args.args[1] == connection_id
            assert retrieve_ex.call_args.args[2] == problem._thread_id
            assert retrieve_ex.call_args.kwargs["for_update"] is True

            save_ex.assert_called_once()

            assert ret_exchange.state == V10CredentialExchange.STATE_ABANDONED

    async def test_receive_problem_report_x(self):
        connection_id = "connection-id"
        problem = CredentialProblemReport(
            description={
                "code": test_module.ProblemReportReason.ISSUANCE_ABANDONED.value,
                "en": "Insufficient privilege",
            }
        )

        with mock.patch.object(
            V10CredentialExchange,
            "retrieve_by_connection_and_thread",
            mock.CoroutineMock(),
        ) as retrieve_ex:
            retrieve_ex.side_effect = test_module.StorageNotFoundError("No such record")

            exch = await self.manager.receive_problem_report(problem, connection_id)
            assert exch is None

    async def test_retrieve_records(self):
        self.cache = InMemoryCache()
        self.profile.context.injector.bind_instance(BaseCache, self.cache)

        for index in range(2):
            exchange_record = V10CredentialExchange(
                connection_id=str(index),
                thread_id=str(1000 + index),
                initiator=V10CredentialExchange.INITIATOR_SELF,
                role=V10CredentialExchange.ROLE_ISSUER,
            )
            async with self.profile.session() as session:
                await exchange_record.save(session)

                for _ in range(2):  # second pass gets from cache
                    ret_ex = (
                        await V10CredentialExchange.retrieve_by_connection_and_thread(
                            session, str(index), str(1000 + index)
                        )
                    )
                    assert ret_ex.connection_id == str(index)
                    assert ret_ex.thread_id == str(1000 + index)



from unittest import TestCase

from ..util import (
    abbr_verkey,
    b58_to_bytes,
    b64_to_str,
    bytes_to_b58,
    bytes_to_b64,
    full_verkey,
    pad,
    set_urlsafe_b64,
    str_to_b64,
    unpad,
)

BYTES = b"\xe0\xa0\xbe"  # chr(2110).encode(), least with + in b64 encoding
STR = "Hello World"  # b64encodes to SGVsbG8gV29ybGQ=


class TestUtil(TestCase):
    def test_b64_urlsafe(self):
        for urlsafe in (False, True):
            CHAR62 = ["+", "-"]
            b64 = bytes_to_b64(BYTES, urlsafe=urlsafe)
            assert CHAR62[urlsafe] in b64

            b64 = set_urlsafe_b64(b64, urlsafe=urlsafe)
            assert CHAR62[urlsafe] in b64
            assert CHAR62[not urlsafe] not in b64

            b64 = set_urlsafe_b64(b64, urlsafe=(not urlsafe))
            assert CHAR62[urlsafe] not in b64
            assert CHAR62[not urlsafe] in b64

    def test_b64_str(self):
        b64 = str_to_b64(STR)
        assert "=" in b64
        assert b64_to_str(b64) == STR

        b64 = str_to_b64(STR, pad=False)
        assert "=" not in b64
        assert b64_to_str(b64) == STR

        b64 = str_to_b64(STR, urlsafe=True)
        assert "=" in b64
        assert b64_to_str(b64, urlsafe=True) == STR
        assert b64_to_str(b64, urlsafe=True, encoding="ascii") == STR

        b64 = str_to_b64(STR, urlsafe=True, pad=False)
        assert "=" not in b64
        assert b64_to_str(b64, urlsafe=True) == STR
        assert b64_to_str(b64, urlsafe=True, encoding="ascii") == STR

    def test_pad(self):
        assert pad("SGVsbG8gV29ybGQ") == "SGVsbG8gV29ybGQ="
        assert unpad("SGVsbG8gV29ybGQ=") == "SGVsbG8gV29ybGQ"

        assert pad("SGVsbG8gV29ybGQx") == "SGVsbG8gV29ybGQx"
        assert unpad("SGVsbG8gV29ybGQx") == "SGVsbG8gV29ybGQx"

    def test_b58(self):
        b58 = bytes_to_b58(BYTES)
        assert b58_to_bytes(b58) == BYTES

    def test_full_abbr_verkey(self):
        did = "N76kAdywAdHCySjgymbZ9t"
        full_vk = "CWBBfFmEVUDbs7rSCeKaFKfaeYXS2dKynAk7e2sCv23b"
        abbr_vk = "~79woMYnyEk6XnQaBA39i57"

        full = full_verkey(did, abbr_vk)
        assert full == full_vk
        assert full == full_verkey(f"did:sov:{did}", abbr_vk)
        assert full_verkey(did, full_vk) == full_vk

        abbr = abbr_verkey(full_vk)
        abbr_did = abbr_verkey(full_vk, did)
        abbr_qdid = abbr_verkey(full_vk, f"did:sov:{did}")
        assert abbr_did == abbr_qdid == abbr == abbr_vk



"""Profile manager for multiple Indy ledger support."""

import logging
from collections import OrderedDict

from ...cache.base import BaseCache
from ...config.injector import BaseInjector, InjectionError
from ...config.provider import BaseProvider
from ...config.settings import BaseSettings
from ...core.profile import Profile
from ...utils.classloader import ClassNotFoundError, DeferLoad
from .base_manager import MultipleLedgerManagerError

LOGGER = logging.getLogger(__name__)


class MultiIndyLedgerManagerProvider(BaseProvider):
    """Multiple Indy ledger support manager provider."""

    MANAGER_TYPES = {
        "single-wallet-askar": (
            DeferLoad(
                "acapy_agent.ledger.multiple_ledger."
                "indy_vdr_manager.MultiIndyVDRLedgerManager"
            )
        ),
    }
    LEDGER_TYPES = {
        "single-wallet-askar": {
            "pool": DeferLoad("acapy_agent.ledger.indy_vdr.IndyVdrLedgerPool"),
            "ledger": DeferLoad("acapy_agent.ledger.indy_vdr.IndyVdrLedger"),
        },
    }

    def __init__(self, root_profile):
        """Initialize the multiple Indy ledger profile manager provider."""
        self._inst = {}
        self.root_profile: Profile = root_profile

    def provide(self, settings: BaseSettings, injector: BaseInjector):
        """Create the multiple Indy ledger manager instance."""

        backend_name = self.root_profile.BACKEND_NAME
        if backend_name in ("askar", "askar-anoncreds"):
            manager_type = "single-wallet-askar"
        else:
            raise MultipleLedgerManagerError(f"Unexpected wallet backend: {backend_name}")

        if manager_type not in self._inst:
            manager_class = self.MANAGER_TYPES.get(manager_type)
            pool_class = self.LEDGER_TYPES[manager_type]["pool"]
            ledger_class = self.LEDGER_TYPES[manager_type]["ledger"]
            LOGGER.info("Create multiple Indy ledger manager: %s", manager_type)
            try:
                indy_vdr_production_ledgers = OrderedDict()
                indy_vdr_non_production_ledgers = OrderedDict()
                ledger_config_list = settings.get_value("ledger.ledger_config_list")
                ledger_endorser_map = {}
                write_ledgers = set()
                for config in ledger_config_list:
                    keepalive = config.get("keepalive")
                    read_only = config.get("read_only")
                    socks_proxy = config.get("socks_proxy")
                    genesis_transactions = config.get("genesis_transactions")
                    cache = injector.inject_or(BaseCache)
                    ledger_id = config.get("id")
                    pool_name = config.get("pool_name")
                    ledger_is_production = config.get("is_production")
                    ledger_is_write = config.get("is_write")
                    ledger_endorser_alias = config.get("endorser_alias")
                    ledger_endorser_did = config.get("endorser_did")
                    ledger_pool = pool_class(
                        pool_name,
                        keepalive=keepalive,
                        cache=cache,
                        genesis_transactions=genesis_transactions,
                        read_only=read_only,
                        socks_proxy=socks_proxy,
                    )
                    ledger_instance = ledger_class(
                        pool=ledger_pool,
                        profile=self.root_profile,
                    )
                    if ledger_is_write:
                        write_ledgers.add(ledger_id)
                    if ledger_is_production:
                        indy_vdr_production_ledgers[ledger_id] = ledger_instance
                    else:
                        indy_vdr_non_production_ledgers[ledger_id] = ledger_instance
                    if ledger_endorser_alias and ledger_endorser_did:
                        ledger_endorser_map[ledger_id] = {
                            "endorser_alias": ledger_endorser_alias,
                            "endorser_did": ledger_endorser_did,
                        }
                self._inst[manager_type] = manager_class(
                    self.root_profile,
                    production_ledgers=indy_vdr_production_ledgers,
                    non_production_ledgers=indy_vdr_non_production_ledgers,
                    writable_ledgers=write_ledgers,
                    endorser_map=ledger_endorser_map,
                )
            except ClassNotFoundError as err:
                raise InjectionError(
                    f"Unknown multiple Indy ledger manager type: {manager_type}"
                ) from err

        return self._inst[manager_type]



"""Handle registration of plugin modules for extending functionality."""

import logging
from collections import OrderedDict
from types import ModuleType
from typing import Optional, Sequence, Set

from ..config.injection_context import InjectionContext
from ..core.event_bus import EventBus
from ..utils.classloader import ClassLoader, ModuleLoadError
from .error import ProtocolDefinitionValidationError
from .goal_code_registry import GoalCodeRegistry
from .protocol_registry import ProtocolRegistry

LOGGER = logging.getLogger(__name__)


class PluginRegistry:
    """Plugin registry for indexing application plugins."""

    def __init__(self, blocklist: Optional[Set[str]] = None):
        """Initialize a `PluginRegistry` instance."""
        self._plugins: OrderedDict[str, ModuleType] = OrderedDict()
        self._blocklist: Set[str] = set(blocklist) if blocklist else set()

    @property
    def plugin_names(self) -> Sequence[str]:
        """Accessor for a list of all plugin modules."""
        return list(self._plugins.keys())

    @property
    def plugins(self) -> Sequence[ModuleType]:
        """Accessor for a list of all plugin modules."""
        return list(self._plugins.values())

    def validate_version(self, version_list, module_name):
        """Validate version dict format."""

        is_list = isinstance(version_list, list)

        # Must be a list
        if not is_list:
            raise ProtocolDefinitionValidationError(
                "Versions definition is not of type list"
            )

        # Must have at least one definition
        if len(version_list) < 1:
            raise ProtocolDefinitionValidationError(
                "Versions list must define at least one version module"
            )

        if not all(isinstance(v, dict) for v in version_list):
            raise ProtocolDefinitionValidationError(
                "Element of versions definition list is not of type dict"
            )

        for version_dict in version_list:
            # Dicts must have correct format
            try:
                if not (
                    isinstance(version_dict["major_version"], int)
                    and isinstance(version_dict["minimum_minor_version"], int)
                    and isinstance(version_dict["current_minor_version"], int)
                    and isinstance(version_dict["path"], str)
                ):
                    raise ProtocolDefinitionValidationError(
                        "Unexpected types in version definition"
                    )
            except KeyError as e:
                raise ProtocolDefinitionValidationError(
                    f"Element of versions definition list is missing an attribute: {e}"
                )

            # Version number cannot be negative
            if (
                version_dict["major_version"] < 0
                or version_dict["minimum_minor_version"] < 0
                or version_dict["current_minor_version"] < 0
            ):
                raise ProtocolDefinitionValidationError(
                    "Version number cannot be negative"
                )

            # Minimum minor version cannot be great than current version
            if (
                version_dict["minimum_minor_version"]
                > version_dict["current_minor_version"]
            ):
                raise ProtocolDefinitionValidationError(
                    "Minimum supported minor version cannot "
                    "be greater than current minor version"
                )

            # There can only be one definition per major version
            major_version = version_dict["major_version"]
            count = 0
            for version_dict_outer in version_list:
                if version_dict_outer["major_version"] == major_version:
                    count += 1
            if count > 1:
                raise ProtocolDefinitionValidationError(
                    "There can only be one definition per major version. "
                    f"Found {count} for major version {major_version}."
                )

            # Specified module must be loadable
            version_path = version_dict["path"]
            mod = ClassLoader.load_module(version_path, module_name)

            if not mod:
                raise ProtocolDefinitionValidationError(
                    f"Version module path is not loadable: {module_name}, {version_path}"
                )

        return True

    def register_plugin(self, module_name: str) -> Optional[ModuleType]:
        """Register a plugin module."""
        if self._is_already_registered(module_name):
            return self._plugins.get(module_name)

        if self._is_blocked(module_name):
            return None

        mod = self._load_module(module_name)
        if not mod:
            LOGGER.error("Module doesn't exist: %s", module_name)
            return None

        if self._is_valid_plugin(mod, module_name):
            self._plugins[module_name] = mod
            LOGGER.debug("Registered plugin: %s", module_name)
            return mod

        LOGGER.warning("Failed to register plugin: %s", module_name)
        return None

    def _is_already_registered(self, module_name: str) -> bool:
        """Check if the plugin is already registered."""
        if module_name in self._plugins:
            LOGGER.debug("Plugin %s is already registered.", module_name)
            return True
        return False

    def _is_blocked(self, module_name: str) -> bool:
        """Check if the plugin is in the blocklist."""
        if module_name in self._blocklist:
            LOGGER.debug("Blocked %s from loading due to blocklist.", module_name)
            return True
        return False

    def _load_module(self, module_name: str) -> Optional[ModuleType]:
        """Load the plugin module using ClassLoader."""
        try:
            mod = ClassLoader.load_module(module_name)
            return mod
        except ModuleLoadError as e:
            LOGGER.error("Error loading plugin module '%s': %s", module_name, e)
        return None

    def _is_valid_plugin(self, mod: ModuleType, module_name: str) -> bool:
        """Validate the plugin based on various criteria."""
        # Check if the plugin has a 'setup' method
        if hasattr(mod, "setup"):
            return True

        # Check for 'routes' or 'message_types' modules
        # This makes an exception for non-protocol modules that contain admin routes
        # and for old-style protocol modules without version support
        routes = ClassLoader.load_module("routes", module_name)
        message_types = ClassLoader.load_module("message_types", module_name)
        if routes or message_types:
            return True

        # Check for 'definition' module with 'versions' attribute
        definition = ClassLoader.load_module("definition", module_name)

        # definition.py must exist in protocol
        if not definition:
            LOGGER.error(
                "Protocol does not include 'definition.py' for module: %s",
                module_name,
            )
            return False

        # definition.py must include versions attribute
        if not hasattr(definition, "versions"):
            LOGGER.error(
                "Protocol definition does not include versions attribute for module: %s",
                module_name,
            )
            return False

        # Validate the 'versions' attribute
        try:
            self.validate_version(definition.versions, module_name)
            return True
        except ProtocolDefinitionValidationError as e:
            LOGGER.error(
                "Protocol versions definition is malformed for module '%s': %s",
                module_name,
                e,
            )
            return False

    def register_package(self, package_name: str) -> Sequence[ModuleType]:
        """Register all modules (sub-packages) under a given package name."""
        LOGGER.debug("Registering package: %s", package_name)
        try:
            module_names = ClassLoader.scan_subpackages(package_name)
        except ModuleLoadError:
            LOGGER.error("Plugin module package not found: %s", package_name)
            module_names = []

        registered_plugins = []
        for module_name in module_names:
            # Skip any module whose last segment is 'tests'
            if module_name.split(".")[-1] == "tests":
                continue

            plugin = self.register_plugin(module_name)
            if plugin:
                registered_plugins.append(plugin)
            else:
                LOGGER.warning(
                    "Failed to register %s under %s", module_name, package_name
                )

        return registered_plugins

    async def init_context(self, context: InjectionContext) -> None:
        """Call plugin setup methods on the current context."""
        LOGGER.debug("Initializing plugin context for %d plugins", len(self._plugins))

        for plugin in self._plugins.values():
            if hasattr(plugin, "setup"):
                await plugin.setup(context)
            else:
                await self.load_protocols(context, plugin)

        # register event handlers for each protocol, if provided
        self.register_protocol_events(context)

    async def load_protocol_version(
        self,
        context: InjectionContext,
        mod: ModuleType,
        version_definition: Optional[dict] = None,
    ) -> None:
        """Load a particular protocol version."""
        protocol_registry = context.inject(ProtocolRegistry)
        goal_code_registry = context.inject(GoalCodeRegistry)

        if hasattr(mod, "MESSAGE_TYPES"):
            protocol_registry.register_message_types(
                mod.MESSAGE_TYPES, version_definition=version_definition
            )

        if hasattr(mod, "CONTROLLERS"):
            protocol_registry.register_controllers(mod.CONTROLLERS)
            goal_code_registry.register_controllers(mod.CONTROLLERS)

    async def load_protocols(self, context: InjectionContext, plugin: ModuleType) -> None:
        """For modules that don't implement setup, register protocols manually."""
        plugin_name = plugin.__name__

        # If this module contains message_types, then assume that
        # this is a valid module of the old style (not versioned)
        try:
            message_types_path = f"{plugin_name}.message_types"
            mod = ClassLoader.load_module(message_types_path)
        except ModuleLoadError as e:
            LOGGER.error("Error loading plugin module message types: %s", e)
            return

        if mod:
            await self.load_protocol_version(context, mod)
        else:
            # Otherwise, try check for definition.py for versioned protocol packages
            try:
                definition_path = f"{plugin_name}.definition"
                definition = ClassLoader.load_module(definition_path)
            except ModuleLoadError as e:
                LOGGER.error("Error loading plugin definition module: %s", e)
                return

            if definition:
                for protocol_version in definition.versions:
                    version_path = (
                        f"{plugin_name}.{protocol_version['path']}.message_types"
                    )
                    try:
                        mod = ClassLoader.load_module(version_path)
                    except ModuleLoadError as e:
                        LOGGER.error(
                            "Error loading plugin module message types from %s: %s",
                            version_path,
                            e,
                        )
                        return

                    if mod:
                        await self.load_protocol_version(context, mod, protocol_version)
                    else:
                        LOGGER.debug("Failed to load %s", version_path)

    async def register_admin_routes(self, app) -> None:
        """Call route registration methods on the current context."""
        LOGGER.debug("Registering admin routes for %d plugins", len(self._plugins))

        for plugin in self._plugins.values():
            plugin_name = plugin.__name__
            mod = None
            definition = ClassLoader.load_module("definition", plugin_name)
            if definition:
                # Load plugin routes that are in a versioned package.
                for plugin_version in definition.versions:
                    version_path = f"{plugin_name}.{plugin_version['path']}.routes"
                    try:
                        mod = ClassLoader.load_module(version_path)
                    except ModuleLoadError as e:
                        LOGGER.error(
                            "Error loading admin routes from %s: %s", version_path, e
                        )
                        continue

                    if mod and hasattr(mod, "register"):
                        await mod.register(app)
            else:
                # Load plugin routes that aren't in a versioned package.
                routes_path = f"{plugin_name}.routes"
                try:
                    mod = ClassLoader.load_module(routes_path)
                except ModuleLoadError as e:
                    LOGGER.error("Error loading admin routes from %s: %s", routes_path, e)
                    continue

                if mod and hasattr(mod, "register"):
                    await mod.register(app)

    def register_protocol_events(self, context: InjectionContext) -> None:
        """Call route register_events methods on the current context."""
        LOGGER.debug("Registering protocol events for %d plugins", len(self._plugins))

        event_bus = context.inject_or(EventBus)
        if not event_bus:
            LOGGER.error("No event bus in context")
            return

        for plugin in self._plugins.values():
            plugin_name = plugin.__name__
            mod = None
            definition = ClassLoader.load_module("definition", plugin_name)
            if definition:
                # Load plugin routes that are in a versioned package.
                for plugin_version in definition.versions:
                    version_path = f"{plugin_name}.{plugin_version['path']}.routes"
                    try:
                        mod = ClassLoader.load_module(version_path)
                    except ModuleLoadError as e:
                        LOGGER.error("Error loading events from %s: %s", version_path, e)
                        continue

                    if mod and hasattr(mod, "register_events"):
                        mod.register_events(event_bus)
            else:
                # Load plugin routes that aren't in a versioned package.
                routes_path = f"{plugin_name}.routes"
                try:
                    mod = ClassLoader.load_module(routes_path)
                except ModuleLoadError as e:
                    LOGGER.error("Error loading events from %s: %s", routes_path, e)
                    continue

                if mod and hasattr(mod, "register_events"):
                    mod.register_events(event_bus)

    def post_process_routes(self, app) -> None:
        """Call route binary file response OpenAPI fixups if applicable."""
        LOGGER.debug("Post-processing routes for %d plugins", len(self._plugins))

        for plugin in self._plugins.values():
            plugin_name = plugin.__name__
            mod = None
            definition = ClassLoader.load_module("definition", plugin_name)
            if definition:
                # Set binary file responses for routes that are in a versioned package.
                for plugin_version in definition.versions:
                    version_path = f"{plugin_name}.{plugin_version['path']}.routes"
                    try:
                        mod = ClassLoader.load_module(version_path)
                    except ModuleLoadError as e:
                        LOGGER.error("Error loading routes from %s: %s", version_path, e)
                        continue

                    if mod and hasattr(mod, "post_process_routes"):
                        mod.post_process_routes(app)
            else:
                # Set binary file responses for routes not in a versioned package.
                routes_path = f"{plugin_name}.routes"
                try:
                    mod = ClassLoader.load_module(routes_path)
                except ModuleLoadError as e:
                    LOGGER.error("Error loading routes from %s: %s", routes_path, e)
                    continue

                if mod and hasattr(mod, "post_process_routes"):
                    mod.post_process_routes(app)

    def __repr__(self) -> str:
        """Return a string representation for this class."""
        return "<{}>".format(self.__class__.__name__)



"""Generic problem report handler."""

from ....messaging.base_handler import BaseHandler, BaseResponder, RequestContext
from .message import ProblemReport


class ProblemReportHandler(BaseHandler):
    """Problem report handler class."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Handle problem report message.

        Args:
            context: Request context
            responder: Responder used to reply

        """
        self._logger.debug("ProblemReportHandler called with context %s", context)
        assert isinstance(context.message, ProblemReport)

        self._logger.info(
            "Received problem report from: %s, %r",
            context.message_receipt.sender_did,
            context.message,
        )

        await context.profile.notify("acapy::problem_report", context.message.serialize())



from ..validation_result import DocumentVerificationResult, ProofResult, PurposeResult

DOC_TEMPLATE = {
    "@context": [
        "https://w3id.org/security/v2",
        {
            "schema": "http://schema.org/",
            "name": "schema:name",
            "homepage": "schema:url",
            "image": "schema:image",
        },
    ],
    "name": "Manu Sporny",
    "homepage": "https://manu.sporny.org/",
    "image": "https://manu.sporny.org/images/manu.png",
}

DOC_SIGNED = {
    "@context": [
        "https://w3id.org/security/v2",
        {
            "schema": "http://schema.org/",
            "name": "schema:name",
            "homepage": "schema:url",
            "image": "schema:image",
        },
    ],
    "name": "Manu Sporny",
    "homepage": "https://manu.sporny.org/",
    "image": "https://manu.sporny.org/images/manu.png",
    "proof": {
        "proofPurpose": "assertionMethod",
        "created": "2019-12-11T03:50:55+00:00",
        "type": "Ed25519Signature2018",
        "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
        "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..FX3xSAN3BxpHnclqtiCKsHa3f6O1pi_fulEoCNs2YQplYBU7lYSdnIm1BoPo_YCw8AS25pOQo1ufW05mXJlxAw",
    },
}

DOC_TEMPLATE_2020 = {
    "@context": [
        "https://w3id.org/security/v2",
        "https://w3id.org/security/suites/ed25519-2020/v1",
        {
            "schema": "http://schema.org/",
            "name": "schema:name",
            "homepage": "schema:url",
            "image": "schema:image",
        },
    ],
    "name": "Manu Sporny",
    "homepage": "https://manu.sporny.org/",
    "image": "https://manu.sporny.org/images/manu.png",
}

DOC_SIGNED_2020 = {
    "@context": [
        "https://w3id.org/security/v2",
        "https://w3id.org/security/suites/ed25519-2020/v1",
        {
            "schema": "http://schema.org/",
            "name": "schema:name",
            "homepage": "schema:url",
            "image": "schema:image",
        },
    ],
    "name": "Manu Sporny",
    "homepage": "https://manu.sporny.org/",
    "image": "https://manu.sporny.org/images/manu.png",
    "proof": {
        "proofPurpose": "assertionMethod",
        "created": "2019-12-11T03:50:55+00:00",
        "type": "Ed25519Signature2020",
        "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
        "proofValue": "z3R1pBptysFeWSTMeNVefAHQoPodXYT5pV1Nb9H2vpj2ucb3SJDSVpTvKNuV176KoXdtSrJLxot9CZgD6QtsCJZfJ",
    },
}

DOC_TEMPLATE_BBS = {
    "@context": [
        "https://w3id.org/security/v2",
        "https://w3id.org/security/bbs/v1",
        {
            "schema": "http://schema.org/",
            "name": "schema:name",
            "homepage": "schema:url",
            "image": "schema:image",
        },
    ],
    "name": "Manu Sporny",
    "homepage": "https://manu.sporny.org/",
    "image": "https://manu.sporny.org/images/manu.png",
}

DOC_FRAME_BBS = {
    "@context": [
        "https://w3id.org/security/v2",
        "https://w3id.org/security/bbs/v1",
        {
            "schema": "http://schema.org/",
            "name": "schema:name",
            "homepage": "schema:url",
            "image": "schema:image",
        },
    ],
    "@explicit": True,
    "name": {},
}

DOC_SIGNED_BBS = {
    "@context": [
        "https://w3id.org/security/v2",
        "https://w3id.org/security/bbs/v1",
        {
            "schema": "http://schema.org/",
            "name": "schema:name",
            "homepage": "schema:url",
            "image": "schema:image",
        },
    ],
    "name": "Manu Sporny",
    "homepage": "https://manu.sporny.org/",
    "image": "https://manu.sporny.org/images/manu.png",
    "proof": {
        "type": "BbsBlsSignature2020",
        "verificationMethod": "did:key:zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa#zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa",
        "created": "2019-12-11T03:50:55",
        "proofPurpose": "assertionMethod",
        "proofValue": "jYdVXtAFahqd9Zp09EEQuALXFDhKTz/GfbfvjksFEzOl4zrk4xprdJo5eRcpKr2URlpAFRZ9Civr4h3++a2/Vk9sLZ5cm4/LWeY2H0PQEIcizdIX83n0LvNZoS1+jeCI2b2C6UDva3CzFyVHHKOZfw==",
    },
}

DOC_DERIVED_BBS = {
    "@context": [
        "https://w3id.org/security/v2",
        "https://w3id.org/security/bbs/v1",
        {
            "schema": "http://schema.org/",
            "name": "schema:name",
            "homepage": "schema:url",
            "image": "schema:image",
        },
    ],
    "id": "urn:bnid:_:c14n0",
    "name": "Manu Sporny",
    "proof": {
        "type": "BbsBlsSignatureProof2020",
        "nonce": "801XIat4aOz6K0VKiZoNml5X1/6g7dprU1fOyhDvGaZz5bOcmlws3SN4+ac7KfDoD0w=",
        "proofValue": "AAcvk/vII8D39ei+76WGZB0HIv8ZBwyyN3A7gE9ziOKgTfdLH+icCrk1AuqIX0wgRjHMgqtsRb9jtO76JK8NcPeBQAVA0PFrf3I3BRKssVkWhg+hgTUOxbTd1lGcqaiHLDlOqFuPSazkT2n8YZ+GgqooLImaGNLOuUuHxQo+jS0bSCXeeOWvarQyYRoehUEyjNd2AAAAdIl2Qtkgcty4dZPrxK7+VpKkkqtrfLMh/ioJlLVp2/n1hfvaljeeekXEmSIoSDII4gAAAAIQRfaHQzTH4OYd2Jk85iMNGiMzPBpWtjfovn+3j0kTkw3Bv6Vh5PZRJtZI4G32c4k/sxFyTlRvmToIF8VUg9OkmJNIeHJ9Re8BA1kcQBM88YgfTDndKjQg8mmV0LU8mvJxD1RM2yKaYJ5dNJDMouEFAAAABCRy/HyI8YYfcV2opKNCz5jzMmQkCIjJNbDYzvf8/6RuVgljJeb7o3MLKWfxE8hyuaa/SQ/sqV+cGsBtM00iL+lbuDByo0s5RE+i71ZBHPGwCQtmAqkvnDdtPwl0HuA9IUqckIe0OVVvxydVzhU37pf4mG6yCaFHOaHQ02jeYqLT",
        "verificationMethod": "did:key:zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa#zUC72Q7XD4PE4CrMiDVXuvZng3sBvMmaGgNeTUJuzavH2BS7ThbHL9FhsZM9QYY5fqAQ4MB8M9oudz3tfuaX36Ajr97QRW7LBt6WWmrtESe6Bs5NYzFtLWEmeVtvRYVAgjFcJSa",
        "proofPurpose": "assertionMethod",
        "created": "2019-12-11T03:50:55",
    },
}

DOC_VERIFIED = DocumentVerificationResult(
    verified=True,
    document={
        "@context": [
            "https://w3id.org/security/v2",
            {
                "schema": "http://schema.org/",
                "name": "schema:name",
                "homepage": "schema:url",
                "image": "schema:image",
            },
        ],
        "name": "Manu Sporny",
        "homepage": "https://manu.sporny.org/",
        "image": "https://manu.sporny.org/images/manu.png",
        "proof": {
            "proofPurpose": "assertionMethod",
            "created": "2019-12-11T03:50:55+00:00",
            "type": "Ed25519Signature2018",
            "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
            "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..FX3xSAN3BxpHnclqtiCKsHa3f6O1pi_fulEoCNs2YQplYBU7lYSdnIm1BoPo_YCw8AS25pOQo1ufW05mXJlxAw",
        },
    },
    results=[
        ProofResult(
            verified=True,
            proof={
                "@context": [
                    "https://w3id.org/security/v2",
                    {
                        "schema": "http://schema.org/",
                        "name": "schema:name",
                        "homepage": "schema:url",
                        "image": "schema:image",
                    },
                ],
                "proofPurpose": "assertionMethod",
                "created": "2019-12-11T03:50:55+00:00",
                "type": "Ed25519Signature2018",
                "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "jws": "eyJhbGciOiAiRWREU0EiLCAiYjY0IjogZmFsc2UsICJjcml0IjogWyJiNjQiXX0..FX3xSAN3BxpHnclqtiCKsHa3f6O1pi_fulEoCNs2YQplYBU7lYSdnIm1BoPo_YCw8AS25pOQo1ufW05mXJlxAw",
            },
            purpose_result=PurposeResult(
                valid=True,
                controller={
                    "@context": "https://w3id.org/security/v2",
                    "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                    "assertionMethod": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "authentication": [
                        {
                            "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "publicKeyBase58": "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
                        }
                    ],
                    "capabilityDelegation": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "capabilityInvocation": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "keyAgreement": [
                        {
                            "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6LSbkodSr6SU2trs8VUgnrnWtSm7BAPG245ggrBmSrxbv1R",
                            "type": "X25519KeyAgreementKey2019",
                            "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "publicKeyBase58": "5dTvYHaNaB7mk7iA9LqCJEHG2dGZQsvoi8WGzDRtYEf",
                        }
                    ],
                    "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                },
            ),
        )
    ],
)

DOC_VERIFIED_2020 = DocumentVerificationResult(
    verified=True,
    document={
        "@context": [
            "https://w3id.org/security/v2",
            "https://w3id.org/security/suites/ed25519-2020/v1",
            {
                "schema": "http://schema.org/",
                "name": "schema:name",
                "homepage": "schema:url",
                "image": "schema:image",
            },
        ],
        "name": "Manu Sporny",
        "homepage": "https://manu.sporny.org/",
        "image": "https://manu.sporny.org/images/manu.png",
        "proof": {
            "proofPurpose": "assertionMethod",
            "created": "2019-12-11T03:50:55+00:00",
            "type": "Ed25519Signature2020",
            "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
            "proofValue": "z3R1pBptysFeWSTMeNVefAHQoPodXYT5pV1Nb9H2vpj2ucb3SJDSVpTvKNuV176KoXdtSrJLxot9CZgD6QtsCJZfJ",
        },
    },
    results=[
        ProofResult(
            verified=True,
            proof={
                "@context": [
                    "https://w3id.org/security/v2",
                    "https://w3id.org/security/suites/ed25519-2020/v1",
                    {
                        "schema": "http://schema.org/",
                        "name": "schema:name",
                        "homepage": "schema:url",
                        "image": "schema:image",
                    },
                ],
                "proofPurpose": "assertionMethod",
                "created": "2019-12-11T03:50:55+00:00",
                "type": "Ed25519Signature2020",
                "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                "proofValue": "z3R1pBptysFeWSTMeNVefAHQoPodXYT5pV1Nb9H2vpj2ucb3SJDSVpTvKNuV176KoXdtSrJLxot9CZgD6QtsCJZfJ",
            },
            purpose_result=PurposeResult(
                valid=True,
                controller={
                    "@context": "https://w3id.org/security/v2",
                    "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                    "assertionMethod": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "authentication": [
                        {
                            "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "type": "Ed25519VerificationKey2018",
                            "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "publicKeyBase58": "3Dn1SJNPaCXcvvJvSbsFWP2xaCjMom3can8CQNhWrTRx",
                        }
                    ],
                    "capabilityDelegation": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "capabilityInvocation": [
                        "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL"
                    ],
                    "keyAgreement": [
                        {
                            "id": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6LSbkodSr6SU2trs8VUgnrnWtSm7BAPG245ggrBmSrxbv1R",
                            "type": "X25519KeyAgreementKey2019",
                            "controller": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                            "publicKeyBase58": "5dTvYHaNaB7mk7iA9LqCJEHG2dGZQsvoi8WGzDRtYEf",
                        }
                    ],
                    "verificationMethod": "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL#z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL",
                },
            ),
        )
    ],
)



BBS_NESTED_VC_FULL_REVEAL_DOCUMENT_MATTR = {
    "@context": [
        "https://www.w3.org/2018/credentials/v1",
        "https://www.w3.org/2018/credentials/examples/v1",
        "https://w3id.org/security/bbs/v1",
    ],
    "type": ["VerifiableCredential", "UniversityDegreeCredential"],
}



"""Manage Aries-Askar profile interaction."""

import asyncio
import logging
import time
from typing import Any, Mapping, Optional
from weakref import ref

from aries_askar import AskarError, Session, Store

from ..cache.base import BaseCache
from ..config.injection_context import InjectionContext
from ..config.provider import ClassProvider
from ..core.error import ProfileError
from ..core.profile import Profile, ProfileManager, ProfileSession
from ..indy.holder import IndyHolder
from ..indy.issuer import IndyIssuer
from ..ledger.base import BaseLedger
from ..ledger.indy_vdr import IndyVdrLedger, IndyVdrLedgerPool
from ..storage.base import BaseStorage, BaseStorageSearch
from ..storage.vc_holder.base import VCHolder
from ..utils.multi_ledger import get_write_ledger_config_for_profile
from ..wallet.base import BaseWallet
from ..wallet.crypto import validate_seed
from .store import AskarOpenStore, AskarStoreConfig

# import traceback


LOGGER = logging.getLogger(__name__)


class AskarAnonCredsProfile(Profile):
    """Provide access to Aries-Askar profile interaction methods."""

    BACKEND_NAME = "askar-anoncreds"
    TEST_PROFILE_NAME = "test-profile"

    def __init__(
        self,
        opened: AskarOpenStore,
        context: Optional[InjectionContext] = None,
        *,
        profile_id: Optional[str] = None,
    ):
        """Create a new AskarProfile instance."""
        super().__init__(
            context=context, name=profile_id or opened.name, created=opened.created
        )
        self.opened = opened
        self.ledger_pool: Optional[IndyVdrLedgerPool] = None
        self.profile_id = profile_id
        self.init_ledger_pool()
        self.bind_providers()

    @property
    def name(self) -> str:
        """Accessor for the profile name."""
        return self.profile_id or self.opened.name

    @property
    def store(self) -> Store:
        """Accessor for the opened Store instance."""
        return self.opened.store

    async def remove(self):
        """Remove the profile."""
        if self.profile_id:
            await self.store.remove_profile(self.profile_id)

    def init_ledger_pool(self):
        """Initialize the ledger pool."""
        if self.settings.get("ledger.disabled"):
            LOGGER.debug("init_ledger_pool: Ledger support is disabled")
            return
        if self.settings.get("ledger.genesis_transactions"):
            pool_name = self.settings.get("ledger.pool_name", "default")
            keepalive = int(self.settings.get("ledger.keepalive", 5))
            read_only = bool(self.settings.get("ledger.read_only", False))
            socks_proxy = self.settings.get("ledger.socks_proxy")
            if read_only:
                LOGGER.warning("Note: setting ledger to read-only mode")
            genesis_transactions = self.settings.get("ledger.genesis_transactions")
            cache = self.context.injector.inject_or(BaseCache)
            self.ledger_pool = IndyVdrLedgerPool(
                pool_name,
                keepalive=keepalive,
                cache=cache,
                genesis_transactions=genesis_transactions,
                read_only=read_only,
                socks_proxy=socks_proxy,
            )

    def bind_providers(self):
        """Initialize the profile-level instance providers."""
        injector = self._context.injector

        injector.bind_provider(
            BaseStorageSearch,
            ClassProvider("acapy_agent.storage.askar.AskarStorageSearch", ref(self)),
        )
        injector.bind_provider(
            VCHolder,
            ClassProvider(
                "acapy_agent.storage.vc_holder.askar.AskarVCHolder",
                ref(self),
            ),
        )
        injector.bind_provider(
            IndyHolder,
            ClassProvider(
                "acapy_agent.indy.credx.holder.IndyCredxHolder",
                ref(self),
            ),
        )
        injector.bind_provider(
            IndyIssuer,
            ClassProvider("acapy_agent.indy.credx.issuer.IndyCredxIssuer", ref(self)),
        )

        ledger_config_list = self.settings.get("ledger.ledger_config_list")
        if ledger_config_list:
            write_ledger_config = get_write_ledger_config_for_profile(
                settings=self.settings
            )
            cache = self.context.injector.inject_or(BaseCache)

            pool_name = write_ledger_config.get("pool_name")
            pool_id = write_ledger_config.get("id")
            ledger_name = pool_name or pool_id
            keepalive = write_ledger_config.get("keepalive")
            read_only = write_ledger_config.get("read_only")
            socks_proxy = write_ledger_config.get("socks_proxy")
            genesis_transactions = write_ledger_config.get("genesis_transactions")

            ledger_pool = IndyVdrLedgerPool(
                name=ledger_name,
                keepalive=keepalive,
                cache=cache,
                genesis_transactions=genesis_transactions,
                read_only=read_only,
                socks_proxy=socks_proxy,
            )
            injector.bind_provider(
                BaseLedger,
                ClassProvider(IndyVdrLedger, ledger_pool, ref(self)),
            )
            self.settings["ledger.write_ledger"] = pool_id
            if (
                "endorser_alias" in write_ledger_config
                and "endorser_did" in write_ledger_config
            ):
                self.settings["endorser.endorser_alias"] = write_ledger_config.get(
                    "endorser_alias"
                )
                self.settings["endorser.endorser_public_did"] = write_ledger_config.get(
                    "endorser_did"
                )
        elif self.ledger_pool:
            injector.bind_provider(
                BaseLedger, ClassProvider(IndyVdrLedger, self.ledger_pool, ref(self))
            )

    def session(
        self, context: Optional[InjectionContext] = None
    ) -> "AskarAnonCredsProfileSession":
        """Start a new interactive session with no transaction support requested."""
        return AskarAnonCredsProfileSession(self, False, context=context)

    def transaction(
        self, context: Optional[InjectionContext] = None
    ) -> "AskarAnonCredsProfileSession":
        """Start a new interactive session with commit and rollback support.

        If the current backend does not support transactions, then commit
        and rollback operations of the session will not have any effect.
        """
        return AskarAnonCredsProfileSession(self, True, context=context)

    async def close(self):
        """Close the profile instance."""
        if self.opened:
            await self.opened.close()
            self.opened = None


class AskarAnonCredsProfileSession(ProfileSession):
    """An active connection to the profile management backend."""

    def __init__(
        self,
        profile: AskarAnonCredsProfile,
        is_txn: bool,
        *,
        context: Optional[InjectionContext] = None,
        settings: Mapping[str, Any] = None,
    ):
        """Create a new AskarAnonCredsProfileSession instance."""
        super().__init__(profile=profile, context=context, settings=settings)
        if is_txn:
            self._opener = self.profile.store.transaction(profile.profile_id)
        else:
            self._opener = self.profile.store.session(profile.profile_id)
        self._profile = profile
        self._handle: Optional[Session] = None
        self._acquire_start: Optional[float] = None
        self._acquire_end: Optional[float] = None

    @property
    def handle(self) -> Session:
        """Accessor for the Session instance."""
        return self._handle

    @property
    def store(self) -> Store:
        """Accessor for the Store instance."""
        return self._profile and self._profile.store

    @property
    def is_transaction(self) -> bool:
        """Check if the session supports commit and rollback operations."""
        if self._handle:
            return self._handle.is_transaction
        return self._opener.is_transaction

    async def _setup(self):
        """Create the session or transaction connection, if needed."""
        self._acquire_start = time.perf_counter()
        try:
            self._handle = await asyncio.wait_for(self._opener, 10)
        except AskarError as err:
            raise ProfileError("Error opening store session") from err
        self._acquire_end = time.perf_counter()
        self._opener = None

        injector = self._context.injector
        injector.bind_provider(
            BaseWallet,
            ClassProvider("acapy_agent.wallet.askar.AskarWallet", ref(self)),
        )
        injector.bind_provider(
            BaseStorage,
            ClassProvider("acapy_agent.storage.askar.AskarStorage", ref(self)),
        )

    async def _teardown(self, commit: Optional[bool] = None):
        """Dispose of the session or transaction connection."""
        if commit:
            try:
                await self._handle.commit()
            except AskarError as err:
                raise ProfileError("Error committing transaction") from err
        if self._handle:
            await self._handle.close()
        self._handle = None
        self._check_duration()

    def _check_duration(self):
        # LOGGER.error(
        #     "release session after %.2f, acquire took %.2f",
        #     end - self._acquire_end,
        #     self._acquire_end - self._acquire_start,
        # )
        # end = time.perf_counter()
        # if end - self._acquire_end > 1.0:
        #     LOGGER.error("Long session")
        #     traceback.print_stack(limit=5)
        pass

    def __del__(self):
        """Delete magic method."""
        if self._handle:
            self._check_duration()


class AskarAnonProfileManager(ProfileManager):
    """Manager for Aries-Askar stores."""

    async def provision(
        self, context: InjectionContext, config: Mapping[str, Any] = None
    ) -> Profile:
        """Provision a new instance of a profile."""
        store_config = AskarStoreConfig(config)
        opened = await store_config.open_store(
            provision=True, in_memory=config.get("test")
        )
        return AskarAnonCredsProfile(opened, context)

    async def open(
        self, context: InjectionContext, config: Mapping[str, Any] = None
    ) -> Profile:
        """Open an instance of an existing profile."""
        store_config = AskarStoreConfig(config)
        opened = await store_config.open_store(
            provision=False, in_memory=config.get("test")
        )
        return AskarAnonCredsProfile(opened, context)

    @classmethod
    async def generate_store_key(self, seed: Optional[str] = None) -> str:
        """Generate a raw store key."""
        return Store.generate_raw_key(validate_seed(seed))



from .bbs_nested_vc_full_reveal_document_mattr import (
    BBS_NESTED_VC_FULL_REVEAL_DOCUMENT_MATTR,
)
from .bbs_nested_vc_mattr import BBS_NESTED_VC_MATTR
from .bbs_nested_vc_reveal_document_mattr import BBS_NESTED_VC_REVEAL_DOCUMENT_MATTR
from .bbs_partial_proof_nested_vc_mattr import BBS_PARTIAL_PROOF_NESTED_VC_MATTR
from .bbs_partial_proof_vc_mattr import BBS_PARTIAL_PROOF_VC_MATTR
from .bbs_proof_nested_vc_mattr import BBS_PROOF_NESTED_VC_MATTR
from .bbs_proof_vc_mattr import BBS_PROOF_VC_MATTR
from .bbs_signed_nested_vc_mattr import BBS_SIGNED_NESTED_VC_MATTR
from .bbs_signed_vc_mattr import BBS_SIGNED_VC_MATTR
from .bbs_vc_mattr import BBS_VC_MATTR
from .bbs_vc_reveal_document_mattr import BBS_VC_REVEAL_DOCUMENT_MATTR
from .test_ld_document import TEST_LD_DOCUMENT
from .test_ld_document_bad_partial_proof_bbs import TEST_LD_DOCUMENT_BAD_PARTIAL_PROOF_BBS
from .test_ld_document_bad_signed_bbs import TEST_LD_DOCUMENT_BAD_SIGNED_BBS
from .test_ld_document_bad_signed_ed25519 import TEST_LD_DOCUMENT_BAD_SIGNED_ED25519
from .test_ld_document_bad_signed_ed25519_2020 import (
    TEST_LD_DOCUMENT_BAD_SIGNED_ED25519_2020,
)
from .test_ld_document_bad_signed_secp256r1_2019 import (
    TEST_LD_DOCUMENT_BAD_SIGNED_P256_2019,
)
from .test_ld_document_partial_proof_bbs import TEST_LD_DOCUMENT_PARTIAL_PROOF_BBS
from .test_ld_document_proof_bbs import TEST_LD_DOCUMENT_PROOF_BBS
from .test_ld_document_reveal import TEST_LD_DOCUMENT_REVEAL
from .test_ld_document_reveal_all import TEST_LD_DOCUMENT_REVEAL_ALL
from .test_ld_document_signed_bbs import TEST_LD_DOCUMENT_SIGNED_BBS
from .test_ld_document_signed_ed25519 import TEST_LD_DOCUMENT_SIGNED_ED25519
from .test_ld_document_signed_ed25519_2020 import TEST_LD_DOCUMENT_SIGNED_ED25519_2020
from .test_ld_document_signed_secp256r1_2019 import TEST_LD_DOCUMENT_SIGNED_P256_2019
from .test_vc_document import TEST_VC_DOCUMENT
from .test_vc_document_did_key_ed25519 import TEST_VC_DOCUMENT_DID_KEY_ED25519
from .test_vc_document_nested import TEST_VC_DOCUMENT_NESTED
from .test_vc_document_nested_partial_proof_bbs import (
    TEST_VC_DOCUMENT_NESTED_PARTIAL_PROOF_BBS,
)
from .test_vc_document_nested_proof_bbs import TEST_VC_DOCUMENT_NESTED_PROOF_BBS
from .test_vc_document_nested_reveal import TEST_VC_DOCUMENT_NESTED_REVEAL
from .test_vc_document_nested_signed_bbs import TEST_VC_DOCUMENT_NESTED_SIGNED_BBS
from .test_vc_document_partial_proof_bbs import TEST_VC_DOCUMENT_PARTIAL_PROOF_BBS
from .test_vc_document_reveal import TEST_VC_DOCUMENT_REVEAL
from .test_vc_document_signed_bbs import TEST_VC_DOCUMENT_SIGNED_BBS
from .test_vc_document_signed_did_key_ed25519 import (
    TEST_VC_DOCUMENT_SIGNED_DID_KEY_ED25519,
)
from .test_vc_document_signed_ed25519 import TEST_VC_DOCUMENT_SIGNED_ED25519
from .test_vc_document_signed_ed25519_2020 import TEST_VC_DOCUMENT_SIGNED_ED25519_2020
from .test_vc_document_signed_secp256r1_2019 import TEST_VC_DOCUMENT_SIGNED_P256_2019

__all__ = [
    "BBS_NESTED_VC_FULL_REVEAL_DOCUMENT_MATTR",
    "BBS_NESTED_VC_MATTR",
    "BBS_NESTED_VC_REVEAL_DOCUMENT_MATTR",
    "BBS_PARTIAL_PROOF_NESTED_VC_MATTR",
    "BBS_PARTIAL_PROOF_VC_MATTR",
    "BBS_PROOF_VC_MATTR",
    "BBS_PROOF_NESTED_VC_MATTR",
    "BBS_SIGNED_VC_MATTR",
    "BBS_SIGNED_NESTED_VC_MATTR",
    "BBS_VC_MATTR",
    "BBS_VC_REVEAL_DOCUMENT_MATTR",
    "TEST_LD_DOCUMENT_BAD_SIGNED_BBS",
    "TEST_LD_DOCUMENT_SIGNED_BBS",
    "TEST_LD_DOCUMENT_BAD_SIGNED_ED25519",
    "TEST_LD_DOCUMENT_BAD_SIGNED_ED25519_2020",
    "TEST_LD_DOCUMENT_BAD_SIGNED_P256_2019",
    "TEST_LD_DOCUMENT_SIGNED_ED25519",
    "TEST_LD_DOCUMENT_SIGNED_ED25519_2020",
    "TEST_LD_DOCUMENT_SIGNED_P256_2019",
    "TEST_LD_DOCUMENT",
    "TEST_LD_DOCUMENT_REVEAL",
    "TEST_LD_DOCUMENT_REVEAL_ALL",
    "TEST_LD_DOCUMENT_PROOF_BBS",
    "TEST_LD_DOCUMENT_PARTIAL_PROOF_BBS",
    "TEST_LD_DOCUMENT_BAD_PARTIAL_PROOF_BBS",
    "TEST_VC_DOCUMENT_NESTED_PARTIAL_PROOF_BBS",
    "TEST_VC_DOCUMENT_NESTED_PROOF_BBS",
    "TEST_VC_DOCUMENT_NESTED_SIGNED_BBS",
    "TEST_VC_DOCUMENT_NESTED",
    "TEST_VC_DOCUMENT_PARTIAL_PROOF_BBS",
    "TEST_VC_DOCUMENT_REVEAL",
    "TEST_VC_DOCUMENT_SIGNED_BBS",
    "TEST_VC_DOCUMENT_SIGNED_ED25519",
    "TEST_VC_DOCUMENT_SIGNED_ED25519_2020",
    "TEST_VC_DOCUMENT_SIGNED_P256_2019",
    "TEST_VC_DOCUMENT",
    "TEST_VC_DOCUMENT_SIGNED_DID_KEY_ED25519",
    "TEST_VC_DOCUMENT_DID_KEY_ED25519",
    "TEST_VC_DOCUMENT_NESTED_REVEAL",
]



from unittest import TestCase, mock

from .....didcomm_prefix import DIDCommPrefix
from ...message_types import MENU_REQUEST, PROTOCOL_PACKAGE
from ..menu_request import MenuRequest


class TestMenuRequest(TestCase):
    def setUp(self):
        self.menu_request = MenuRequest()

    def test_init(self):
        """Test initialization."""
        pass

    def test_type(self):
        """Test type."""
        assert self.menu_request._type == DIDCommPrefix.qualify_current(MENU_REQUEST)

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.menu_request.MenuRequestSchema.load")
    def test_deserialize(self, mock_menu_request_schema_load):
        """
        Test deserialization.
        """
        obj = {"obj": "obj"}

        request = MenuRequest.deserialize(obj)
        mock_menu_request_schema_load.assert_called_once_with(obj)

        assert request is mock_menu_request_schema_load.return_value

    @mock.patch(f"{PROTOCOL_PACKAGE}.messages.menu_request.MenuRequestSchema.dump")
    def test_serialize(self, mock_menu_request_schema_dump):
        """
        Test serialization.
        """
        request_dict = self.menu_request.serialize()
        mock_menu_request_schema_dump.assert_called_once_with(self.menu_request)
        assert request_dict is mock_menu_request_schema_dump.return_value

    def test_make_model(self):
        data = self.menu_request.serialize()
        model_instance = MenuRequest.deserialize(data)
        assert isinstance(model_instance, type(self.menu_request))



"""Contains the functions needed to produce and verify a json-ld signature.

This file was ported from
https://github.com/transmute-industries/Ed25519Signature2018/blob/master/src/createVerifyData/index.js
"""

import datetime
import hashlib

from pyld import jsonld

from ...vc.ld_proofs import DocumentLoader
from .error import (
    DroppedAttributeError,
    MissingVerificationMethodError,
    SignatureTypeError,
)


def _canonize(data: dict, document_loader: DocumentLoader | None = None) -> dict:
    return jsonld.normalize(
        data,
        {
            "algorithm": "URDNA2015",
            "format": "application/n-quads",
            **{opt: document_loader for opt in ["documentLoader"] if document_loader},
        },
    )


def _sha256(data: str) -> str:
    return hashlib.sha256(data.encode("utf-8")).hexdigest()


def _canonize_signature_options(
    signature_options: dict, document_loader: DocumentLoader | None = None
) -> dict:
    _signature_options = {**signature_options, "@context": "https://w3id.org/security/v2"}
    _signature_options.pop("jws", None)
    _signature_options.pop("signatureValue", None)
    _signature_options.pop("proofValue", None)
    return _canonize(_signature_options, document_loader)


def _canonize_document(doc: dict, document_loader: DocumentLoader | None = None) -> dict:
    _doc = {**doc}
    _doc.pop("proof", None)
    return _canonize(_doc, document_loader)


def _created_at() -> str:
    """Creation Timestamp."""

    stamp = datetime.datetime.now(datetime.timezone.utc)
    return stamp.strftime("%Y-%m-%dT%H:%M:%SZ")


def create_verify_data(
    data: dict, signature_options: dict, document_loader: DocumentLoader | None = None
) -> tuple[dict, str]:
    """Encapsulate process of constructing string used during sign and verify."""

    signature_options["type"] = signature_options.get("type", "Ed25519Signature2018")
    type_ = signature_options.get("type")
    if type_ != "Ed25519Signature2018":
        raise SignatureTypeError(f"invalid signature type {type_}.")

    if not signature_options.get("verificationMethod"):
        raise MissingVerificationMethodError(
            "signature_options.verificationMethod is required"
        )

    signature_options["created"] = signature_options.get("created", _created_at())
    [expanded] = jsonld.expand(
        data,
        options={
            **{opt: document_loader for opt in ["documentLoader"] if document_loader}
        },
    )
    framed = jsonld.compact(
        expanded,
        "https://w3id.org/security/v2",
        options={
            "skipExpansion": True,
            **{opt: document_loader for opt in ["documentLoader"] if document_loader},
        },
    )

    # Detect any dropped attributes during the expand/contract step.
    if len(data) > len(framed):
        # > check indicates dropped attrs < is a different error
        # attempt to collect error report data
        for_diff = jsonld.compact(
            expanded,
            data.get("@context"),
            options={
                **{opt: document_loader for opt in ["documentLoader"] if document_loader}
            },
        )
        dropped = set(data.keys()) - set(for_diff.keys())
        raise DroppedAttributeError(
            f"{dropped} attributes dropped. Provide definitions in context to correct."
        )
    # Check proof for dropped attributes
    attr = [
        ("proof", "proof"),
        ("credentialSubject", "https://www.w3.org/2018/credentials#credentialSubject"),
    ]
    data_context = data.get("@context")
    for mapping in attr:
        data_attribute = data.get(mapping[0], {})
        frame_attribute = framed.get(mapping[1], {})
        if len(data_attribute) > len(frame_attribute):
            for_diff = jsonld.compact(
                expanded,
                data_context,
                options={
                    **{
                        opt: document_loader
                        for opt in ["documentLoader"]
                        if document_loader
                    }
                },
            )
            for_diff_attribute = for_diff.get(mapping[1], {})
            dropped = set(data_attribute.keys()) - set(for_diff_attribute.keys())
            raise DroppedAttributeError(
                f"in {mapping[0]}, {dropped} attributes dropped. "
                "Provide definitions in context to correct."
            )

    canonized_signature_options = _canonize_signature_options(
        signature_options, document_loader
    )
    hash_of_canonized_signature_options = _sha256(canonized_signature_options)
    canonized_document = _canonize_document(framed, document_loader)
    hash_of_canonized_document = _sha256(canonized_document)

    return (framed, hash_of_canonized_signature_options + hash_of_canonized_document)



"""Admin server routes."""

import asyncio
import re

from aiohttp import web
from aiohttp_apispec import docs, response_schema
from marshmallow import fields

from ..core.plugin_registry import PluginRegistry
from ..messaging.models.openapi import OpenAPISchema
from ..utils.stats import Collector
from ..version import __version__
from .decorators.auth import admin_authentication


class AdminModulesSchema(OpenAPISchema):
    """Schema for the modules endpoint."""

    result = fields.List(
        fields.Str(metadata={"description": "admin module"}),
        metadata={"description": "List of admin modules"},
    )


class AdminConfigSchema(OpenAPISchema):
    """Schema for the config endpoint."""

    config = fields.Dict(
        required=True, metadata={"description": "Configuration settings"}
    )


class AdminStatusSchema(OpenAPISchema):
    """Schema for the status endpoint."""

    version = fields.Str(metadata={"description": "Version code"})
    label = fields.Str(allow_none=True, metadata={"description": "Default label"})
    timing = fields.Dict(required=False, metadata={"description": "Timing results"})
    conductor = fields.Dict(
        required=False, metadata={"description": "Conductor statistics"}
    )


class AdminResetSchema(OpenAPISchema):
    """Schema for the reset endpoint."""


class AdminStatusLivelinessSchema(OpenAPISchema):
    """Schema for the liveliness endpoint."""

    alive = fields.Boolean(metadata={"description": "Liveliness status", "example": True})


class AdminStatusReadinessSchema(OpenAPISchema):
    """Schema for the readiness endpoint."""

    ready = fields.Boolean(metadata={"description": "Readiness status", "example": True})


class AdminShutdownSchema(OpenAPISchema):
    """Response schema for admin Module."""


@docs(tags=["server"], summary="Fetch the list of loaded plugins")
@response_schema(AdminModulesSchema(), 200, description="")
@admin_authentication
async def plugins_handler(request: web.BaseRequest):
    """Request handler for the loaded plugins list.

    Args:
        request: aiohttp request object

    Returns:
        The module list response

    """
    registry = request.app["context"].inject_or(PluginRegistry)
    plugins = registry and sorted(registry.plugin_names) or []
    return web.json_response({"result": plugins})


@docs(tags=["server"], summary="Fetch the server configuration")
@response_schema(AdminConfigSchema(), 200, description="")
@admin_authentication
async def config_handler(request: web.BaseRequest):
    """Request handler for the server configuration.

    Args:
        request: aiohttp request object

    Returns:
        The web response

    """
    config = {
        k: (
            request.app["context"].settings[k]
            if (isinstance(request.app["context"].settings[k], (str, int)))
            else request.app["context"].settings[k].copy()
        )
        for k in request.app["context"].settings
        if k
        not in [
            "admin.admin_api_key",
            "multitenant.jwt_secret",
            "wallet.key",
            "wallet.rekey",
            "wallet.seed",
            "wallet.storage_creds",
        ]
    }
    for index in range(len(config.get("admin.webhook_urls", []))):
        config["admin.webhook_urls"][index] = re.sub(
            r"#.*",
            "",
            config["admin.webhook_urls"][index],
        )

    return web.json_response({"config": config})


@docs(tags=["server"], summary="Fetch the server status")
@response_schema(AdminStatusSchema(), 200, description="")
@admin_authentication
async def status_handler(request: web.BaseRequest):
    """Request handler for the server status information.

    Args:
        request: aiohttp request object

    Returns:
        The web response

    """
    status = {"version": __version__}
    status["label"] = request.app["context"].settings.get("default_label")
    collector = request.app["context"].inject_or(Collector)
    if collector:
        status["timing"] = collector.results
    if request.app["conductor_stats"]:
        status["conductor"] = await request.app["conductor_stats"]()
    return web.json_response(status)


@docs(tags=["server"], summary="Reset statistics")
@response_schema(AdminResetSchema(), 200, description="")
@admin_authentication
async def status_reset_handler(request: web.BaseRequest):
    """Request handler for resetting the timing statistics.

    Args:
        request: aiohttp request object

    Returns:
        The web response

    """
    collector = request.app["context"].inject_or(Collector)
    if collector:
        collector.reset()
    return web.json_response({})


async def redirect_handler(request: web.BaseRequest):
    """Perform redirect to documentation."""
    raise web.HTTPFound("/api/doc")


@docs(tags=["server"], summary="Liveliness check")
@response_schema(AdminStatusLivelinessSchema(), 200, description="")
async def liveliness_handler(request: web.BaseRequest):
    """Request handler for liveliness check.

    Args:
        request: aiohttp request object

    Returns:
        The web response, always indicating True

    """
    app_live = request.app._state["alive"]
    if app_live:
        return web.json_response({"alive": app_live})
    else:
        raise web.HTTPServiceUnavailable(reason="Service not available")


@docs(tags=["server"], summary="Readiness check")
@response_schema(AdminStatusReadinessSchema(), 200, description="")
async def readiness_handler(request: web.BaseRequest):
    """Request handler for liveliness check.

    Args:
        request: aiohttp request object

    Returns:
        The web response, indicating readiness for further calls

    """
    app_ready = request.app._state["ready"] and request.app._state["alive"]
    if app_ready:
        return web.json_response({"ready": app_ready})
    else:
        raise web.HTTPServiceUnavailable(reason="Service not ready")


@docs(tags=["server"], summary="Shut down server")
@response_schema(AdminShutdownSchema(), description="")
@admin_authentication
async def shutdown_handler(request: web.BaseRequest):
    """Request handler for server shutdown.

    Args:
        request: aiohttp request object

    Returns:
        The web response (empty production)

    """
    request.app._state["ready"] = False
    loop = asyncio.get_event_loop()
    asyncio.ensure_future(request.app["conductor_stop"](), loop=loop)

    return web.json_response({})



"""Message type identifiers for Feature Discovery."""

from ...didcomm_prefix import DIDCommPrefix

SPEC_URI = (
    "https://github.com/decentralized-identity/aries-rfcs/tree/"
    "b3a3942ef052039e73cd23d847f42947f8287da2/features/0557-discover-features-v2"
)

# Message types
DISCLOSURES = "discover-features/2.0/disclosures"
QUERIES = "discover-features/2.0/queries"

PROTOCOL_PACKAGE = "acapy_agent.protocols.discovery.v2_0"

MESSAGE_TYPES = DIDCommPrefix.qualify_all(
    {
        DISCLOSURES: f"{PROTOCOL_PACKAGE}.messages.disclosures.Disclosures",
        QUERIES: f"{PROTOCOL_PACKAGE}.messages.queries.Queries",
    }
)



"""Utilities for dealing with Indy conventions."""

from os import getenv, makedirs, urandom
from os.path import isdir, join
from pathlib import Path
from platform import system
from typing import Optional

from aiohttp import web

from .error_messages import ANONCREDS_PROFILE_REQUIRED_MSG


async def generate_pr_nonce() -> str:
    """Generate a nonce for a proof request."""
    # equivalent to indy.anoncreds.generate_nonce
    return str(int.from_bytes(urandom(10), "big"))


def indy_client_dir(subpath: Optional[str] = None, create: bool = False) -> str:
    """Return '/'-terminated subdirectory of indy-client directory.

    Args:
        subpath: subpath within indy-client structure
        create: whether to create subdirectory if absent
    """

    home = Path.home()
    target_dir = join(
        home,
        (
            "Documents"
            if isdir(join(home, "Documents"))
            else getenv("EXTERNAL_STORAGE", "")
            if system() == "Linux"
            else ""
        ),
        ".indy_client",
        subpath if subpath else "",
        "",  # set trailing separator
    )
    if create:
        makedirs(target_dir, exist_ok=True)

    return target_dir


def handle_value_error(e: ValueError) -> None:
    """Handle ValueError message as web response type."""
    if ANONCREDS_PROFILE_REQUIRED_MSG in str(e):
        raise web.HTTPForbidden(reason=str(e)) from e
    raise web.HTTPInternalServerError() from e



"""Test keylist update message."""

from unittest import TestCase

from ...message_types import KEYLIST_UPDATE
from ..inner.keylist_update_rule import KeylistUpdateRule
from ..keylist_update import KeylistUpdate, KeylistUpdateSchema
from . import MessageTest


class TestKeylistUpdate(MessageTest, TestCase):
    """Test keylist update message."""

    TYPE = KEYLIST_UPDATE
    CLASS = KeylistUpdate
    SCHEMA = KeylistUpdateSchema
    VALUES = {
        "updates": [
            KeylistUpdateRule(
                "did:key:z6Mkgg342Ycpuk263R9d8Aq6MUaxPn1DDeHyGo38EefXmgDL", "add"
            )
        ]
    }



"""Message type identifiers for Transactions."""

from ...didcomm_prefix import DIDCommPrefix

# Message types

TRANSACTION_REQUEST = "transactions/1.0/request"
ENDORSED_TRANSACTION_RESPONSE = "transactions/1.0/endorse"
REFUSED_TRANSACTION_RESPONSE = "transactions/1.0/refuse"
CANCEL_TRANSACTION = "transactions/1.0/cancel"
TRANSACTION_RESEND = "transactions/1.0/resend"
TRANSACTION_JOB_TO_SEND = "transactions/1.0/transaction_my_job"
TRANSACTION_ACKNOWLEDGEMENT = "transactions/1.0/ack"
ATTACHED_MESSAGE = "transactions/1.0/message"
PROTOCOL_PACKAGE = "acapy_agent.protocols.endorse_transaction.v1_0"


MESSAGE_TYPES = DIDCommPrefix.qualify_all(
    {
        TRANSACTION_REQUEST: (
            f"{PROTOCOL_PACKAGE}.messages.transaction_request.TransactionRequest"
        ),
        ENDORSED_TRANSACTION_RESPONSE: (
            f"{PROTOCOL_PACKAGE}.messages.endorsed_transaction_response"
            ".EndorsedTransactionResponse"
        ),
        REFUSED_TRANSACTION_RESPONSE: (
            f"{PROTOCOL_PACKAGE}.messages.refused_transaction_response"
            ".RefusedTransactionResponse"
        ),
        CANCEL_TRANSACTION: (
            f"{PROTOCOL_PACKAGE}.messages.cancel_transaction.CancelTransaction"
        ),
        TRANSACTION_RESEND: (
            f"{PROTOCOL_PACKAGE}.messages.transaction_resend.TransactionResend"
        ),
        TRANSACTION_JOB_TO_SEND: (
            f"{PROTOCOL_PACKAGE}.messages.transaction_job_to_send.TransactionJobToSend"
        ),
        TRANSACTION_ACKNOWLEDGEMENT: (
            f"{PROTOCOL_PACKAGE}.messages.transaction_acknowledgement"
            ".TransactionAcknowledgement"
        ),
    }
)

CONTROLLERS = DIDCommPrefix.qualify_all(
    {"transactions/1.0": f"{PROTOCOL_PACKAGE}.controller.Controller"}
)



"""Presentation request message handler."""

from .....anoncreds.holder import AnonCredsHolderError
from .....core.oob_processor import OobMessageProcessor
from .....indy.holder import IndyHolderError
from .....ledger.error import LedgerError
from .....messaging.base_handler import BaseHandler, HandlerException
from .....messaging.models.base import BaseModelError
from .....messaging.request_context import RequestContext
from .....messaging.responder import BaseResponder
from .....storage.error import StorageError, StorageNotFoundError
from .....utils.tracing import get_timer, trace_event
from .....wallet.error import WalletNotFoundError
from .. import problem_report_for_record
from ..formats.handler import V20PresFormatHandlerError
from ..manager import V20PresManager
from ..messages.pres_problem_report import ProblemReportReason
from ..messages.pres_request import V20PresRequest
from ..models.pres_exchange import V20PresExRecord


class V20PresRequestHandler(BaseHandler):
    """Message handler class for v2.0 presentation requests."""

    async def handle(self, context: RequestContext, responder: BaseResponder):
        """Message handler logic for v2.0 presentation requests.

        Args:
            context: request context
            responder: responder callback

        """
        r_time = get_timer()

        self._logger.debug("V20PresRequestHandler called with context %s", context)
        assert isinstance(context.message, V20PresRequest)
        self._logger.debug(
            "Received v2.0 presentation request message: %s",
            context.message.serialize(as_string=True),
        )

        # If connection is present it must be ready for use
        if context.connection_record and not context.connection_ready:
            raise HandlerException("Connection used for presentation request not ready")

        # Find associated oob record
        oob_processor = context.inject(OobMessageProcessor)
        oob_record = await oob_processor.find_oob_record_for_inbound_message(context)

        # Either connection or oob context must be present
        if not context.connection_record and not oob_record:
            raise HandlerException(
                "No connection or associated connectionless exchange found for"
                " presentation request"
            )

        connection_id = (
            context.connection_record.connection_id if context.connection_record else None
        )

        profile = context.profile
        pres_manager = V20PresManager(profile)

        # Get pres ex record (holder initiated via proposal)
        # or create it (verifier sent request first)
        try:
            async with profile.session() as session:
                pres_ex_record = await V20PresExRecord.retrieve_by_tag_filter(
                    session,
                    {"thread_id": context.message._thread_id},
                    {
                        "connection_id": connection_id,
                        "role": V20PresExRecord.ROLE_PROVER,
                    },
                )  # holder initiated via proposal
            pres_ex_record.pres_request = context.message
        except StorageNotFoundError:
            # verifier sent this request free of any proposal
            pres_ex_record = V20PresExRecord(
                connection_id=connection_id,
                thread_id=context.message._thread_id,
                initiator=V20PresExRecord.INITIATOR_EXTERNAL,
                role=V20PresExRecord.ROLE_PROVER,
                pres_request=context.message,
                auto_present=context.settings.get(
                    "debug.auto_respond_presentation_request"
                ),
                trace=(context.message._trace is not None),
                auto_remove=not profile.settings.get("preserve_exchange_records"),
            )

        pres_ex_record = await pres_manager.receive_pres_request(
            pres_ex_record
        )  # mgr only saves record: on exception, saving state err is hopeless

        r_time = trace_event(
            context.settings,
            context.message,
            outcome="V20PresRequestHandler.handle.END",
            perf_counter=r_time,
        )

        # If auto_present is enabled, respond immediately with presentation
        if pres_ex_record.auto_present:
            pres_message = None
            try:
                (pres_ex_record, pres_message) = await pres_manager.create_pres(
                    pres_ex_record=pres_ex_record,
                    comment=(
                        f"auto-presented for proof requests"
                        f", pres_ex_record: {pres_ex_record.pres_ex_id}"
                    ),
                )
                await responder.send_reply(pres_message)
            except (
                BaseModelError,
                AnonCredsHolderError,
                IndyHolderError,
                LedgerError,
                StorageError,
                WalletNotFoundError,
                V20PresFormatHandlerError,
            ) as err:
                self._logger.exception(err)
                if pres_ex_record:
                    async with profile.session() as session:
                        await pres_ex_record.save_error_state(
                            session,
                            reason=err.roll_up,  # us: be specific
                        )
                    await responder.send_reply(
                        problem_report_for_record(
                            pres_ex_record,
                            ProblemReportReason.ABANDONED.value,  # them: be vague
                        )
                    )
            trace_event(
                context.settings,
                pres_message,
                outcome="V20PresRequestHandler.handle.PRESENT",
                perf_counter=r_time,
            )



from unittest import TestCase

from ..proof_request import IndyProofRequest

TEST_DID = "LjgpST2rjsoxYegQDRm7EL"
SCHEMA_NAME = "preferences"
SCHEMA_TXN = 12
SCHEMA_ID = f"{TEST_DID}:2:{SCHEMA_NAME}:1.0"
CRED_DEF_ID = f"{TEST_DID}:3:CL:12:default"
REV_REG_ID = f"{TEST_DID}:4:{CRED_DEF_ID}:CL_ACCUM:0"

PROOF_REQ = {
    "nonce": "123432421212",
    "name": "proof_req_1",
    "version": "0.1",
    "requested_attributes": {
        "attr1_referent": {
            "name": "name",
            "restrictions": [
                {
                    "schema_id": SCHEMA_ID,
                    "schema_issuer_did": TEST_DID,
                    "schema_name": SCHEMA_NAME,
                    "schema_version": "1.0",
                    "issuer_did": TEST_DID,
                    "cred_def_id": CRED_DEF_ID,
                }
            ],
            "non_revoked": {  # overrides proof-level spec
                "from": 1234567890,
                "to": 1234567890,
            },
        }
    },
    "requested_predicates": {
        "predicate1_referent": {
            "name": "age",
            "p_type": ">=",
            "p_value": 18,
            "restrictions": [
                {
                    "schema_id": SCHEMA_ID,
                    "schema_issuer_did": TEST_DID,
                    "schema_name": SCHEMA_NAME,
                    "schema_version": "1.0",
                    "issuer_did": TEST_DID,
                    "cred_def_id": CRED_DEF_ID,
                }
            ],
            "non_revoked": {  # overrides proof-level spec
                "from": 1234567890,
                "to": 1234567890,
            },
        }
    },
    "non_revoked": {"from": 1584704048, "to": 1584704048},
}


class TestIndyProofReq(TestCase):
    """Test indy proof req."""

    def test_serde(self):
        """Test de/serialization."""

        proof_req = IndyProofRequest.deserialize(PROOF_REQ)
        assert type(proof_req) is IndyProofRequest

        ser = proof_req.serialize()
        assert ser == PROOF_REQ

        deser = IndyProofRequest.deserialize(ser)
        reser = deser.serialize()
        assert ser == reser

        obj = IndyProofRequest(**PROOF_REQ)
        ser = obj.serialize()
        assert ser == reser

        obj2 = IndyProofRequest(
            nonce=obj.nonce,
            name=obj.name,
            version=obj.version,
            requested_attributes=obj.requested_attributes,
            requested_predicates=obj.requested_predicates,
            non_revoked=obj.non_revoked,
        )
        assert ser.items() <= obj2.serialize().items()



"""Base Registry."""

from abc import ABC, abstractmethod
from typing import Generic, Optional, Pattern, Sequence, TypeVar

from ..config.injection_context import InjectionContext
from ..core.error import BaseError
from ..core.profile import Profile
from .models.credential_definition import CredDef, CredDefResult, GetCredDefResult
from .models.revocation import (
    GetRevListResult,
    GetRevRegDefResult,
    RevList,
    RevListResult,
    RevRegDef,
    RevRegDefResult,
)
from .models.schema import AnonCredsSchema, GetSchemaResult, SchemaResult
from .models.schema_info import AnonCredsSchemaInfo

T = TypeVar("T")


class BaseAnonCredsError(BaseError):
    """Base error class for AnonCreds."""


class AnonCredsObjectNotFound(BaseAnonCredsError):
    """Raised when object is not found in resolver."""

    def __init__(
        self, message: Optional[str] = None, resolution_metadata: Optional[dict] = None
    ):
        """Constructor."""
        super().__init__(message, resolution_metadata)
        self.resolution_metadata = resolution_metadata or {}


class AnonCredsRegistrationError(BaseAnonCredsError):
    """Raised when registering an AnonCreds object fails."""


class AnonCredsObjectAlreadyExists(AnonCredsRegistrationError, Generic[T]):
    """Raised when an AnonCreds object already exists."""

    def __init__(
        self,
        message: str,
        obj_id: str,
        obj: Optional[T] = None,
        *args,
        **kwargs,
    ):
        """Initialize an instance.

        Args:
            message (str): The message associated with the instance.
            obj_id (str): The ID of the object.
            obj (T, optional): The generic object associated with the instance.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__(message, obj_id, obj, *args, **kwargs)
        self._message = message
        self.obj_id = obj_id
        self.obj = obj

    @property
    def message(self) -> str:
        """Message."""
        return f"{self._message}: {self.obj_id}, {self.obj}"


class AnonCredsSchemaAlreadyExists(AnonCredsObjectAlreadyExists[AnonCredsSchema]):
    """Raised when a schema already exists."""

    @property
    def schema_id(self) -> str:
        """Get Schema Id."""
        return self.obj_id

    @property
    def schema(self) -> AnonCredsSchema:
        """Get Schema."""
        return self.obj


class AnonCredsResolutionError(BaseAnonCredsError):
    """Raised when resolving an AnonCreds object fails."""


class BaseAnonCredsHandler(ABC):
    """Base Anon Creds Handler."""

    @property
    @abstractmethod
    def supported_identifiers_regex(self) -> Pattern:
        """Regex to match supported identifiers."""

    async def supports(self, identifier: str) -> bool:
        """Determine whether this registry supports the given identifier."""
        return bool(self.supported_identifiers_regex.match(identifier))

    @abstractmethod
    async def setup(self, context: InjectionContext):
        """Class Setup method."""


class BaseAnonCredsResolver(BaseAnonCredsHandler):
    """Base Anon Creds Resolver."""

    @abstractmethod
    async def get_schema(self, profile: Profile, schema_id: str) -> GetSchemaResult:
        """Get a schema from the registry."""

    @abstractmethod
    async def get_credential_definition(
        self, profile: Profile, credential_definition_id: str
    ) -> GetCredDefResult:
        """Get a credential definition from the registry."""

    @abstractmethod
    async def get_revocation_registry_definition(
        self, profile: Profile, revocation_registry_id: str
    ) -> GetRevRegDefResult:
        """Get a revocation registry definition from the registry."""

    @abstractmethod
    async def get_revocation_list(
        self,
        profile: Profile,
        revocation_registry_id: str,
        timestamp_from: Optional[int] = 0,
        timestamp_to: Optional[int] = None,
    ) -> GetRevListResult:
        """Get a revocation list from the registry."""

    @abstractmethod
    async def get_schema_info_by_id(
        self, profile: Profile, schema_id: str
    ) -> AnonCredsSchemaInfo:
        """Get a schema info from the registry."""


class BaseAnonCredsRegistrar(BaseAnonCredsHandler):
    """Base AnonCreds Registrar."""

    @abstractmethod
    async def register_schema(
        self,
        profile: Profile,
        schema: AnonCredsSchema,
        options: Optional[dict] = None,
    ) -> SchemaResult:
        """Register a schema on the registry."""

    @abstractmethod
    async def register_credential_definition(
        self,
        profile: Profile,
        schema: GetSchemaResult,
        credential_definition: CredDef,
        options: Optional[dict] = None,
    ) -> CredDefResult:
        """Register a credential definition on the registry."""

    @abstractmethod
    async def register_revocation_registry_definition(
        self,
        profile: Profile,
        revocation_registry_definition: RevRegDef,
        options: Optional[dict] = None,
    ) -> RevRegDefResult:
        """Register a revocation registry definition on the registry."""

    @abstractmethod
    async def register_revocation_list(
        self,
        profile: Profile,
        rev_reg_def: RevRegDef,
        rev_list: RevList,
        options: Optional[dict] = None,
    ) -> RevListResult:
        """Register a revocation list on the registry."""

    @abstractmethod
    async def update_revocation_list(
        self,
        profile: Profile,
        rev_reg_def: RevRegDef,
        prev_list: RevList,
        curr_list: RevList,
        revoked: Sequence[int],
        options: Optional[dict] = None,
    ) -> RevListResult:
        """Update a revocation list on the registry."""
